#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2019 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>Custom Font Labels'
#
# The first option creates new letter faces from over-sized visible layer/s
# The second option clears the contents of the old character layer before copying and pasting the new letter face
# The third and fourth options set the mode (legacy and new) to be used when copying and pasting to generate the new font face
# The fifth option removes non-font character layers from the image - apart from any layer name containing the word 'legend'
# The sixth option will re-position all layers to the origin of their parent image (Option 4 does this too)
# The seventh option will re-order the layers in the custom font file according to the following order
#		space.png (or  .ping)
#		Capital letters (A-Z)
#		Lower case letters (a-z)
#		Numbers (0-9)
#		Basic maths characters
# 		Punctuation characters
#		Other keyboard characters
#		Special characters (Pound Sterling sign, Copyright Sign, Registered Trademark sign)
# You can change this order by altering the appropriate lines (currently beginning at line no 130) of the plug-in
# The eighth option will display a message box listing all the characters (ie layers) defined in the custom font
# The ninth option defines the string to be used to append to single letter layer and/or path names
# The tenth option will add .png to all the layers with one character names ie  A to A.png
# The eleventh option will add .png to all the paths with one character names
# The twelfth option will combine the characters from the xcf custom font file selected, with the file selection tool below, with the current characters
# The thirteenth option is the xcf file selection tool
# The fourteenth option trims pixels from the character

# Version 0.01 Fixed adding .png to single letter layers and paths with special characters ie unicode characters
# Version 1.02 Adds an option to create a new face for a custom font from additional/decorative visible layers
# Version 1.03 Re-position space layer, if one exists, along with other letter layers and adds the option to remove non-font-character layers from the image
# Version 1.04 Trim and feather pixels from letter faces to remove halo effect
# Version 1.05 Fixed bug where character layer is not at image origin (and so doesn't copy new face correctly): Adds option to use different modes when copying from visible layers to font face
# Version 1.06 Added option to re-position all layers to the origin of their parent image ie with x and y co-ordinates set to zero: Re-organized dialog and filter structure so that all options can be set without any issues
# Version 1.07 Added option to clear the existing letter face before copying: Re-ordered flow of filter (yet again); Added options to combine two custom font files
# Version 1.08 Fixed potential bug when layer is bigger than image
#              Changed filter location
# Version 1.09 Added mode choice when combining two custom fonts
# Version 1.10 Added fixed position option for creating new characters from a visible layer
# Version 1.11 Adds an option to trim pixels from each character; changed plug-in name to Custom Font Layer Utilities
# Version 1.12 Adds an option to add lower-case characters to the custom font by copying and scaling, as a percentage, the upper-case characters (when the custom font has a path called base_line.png)
# Version 1.13 Adds an option to sandwich a character between a top layer and a bottom layer - optionally setting the layer size to the full width of the font image size
# Version 1.15 Return trim option inadvertently removed, add lower case now works with part fonts, layers containing the word "Legend" no longer removed from stack
# Version 1.16 Adds options for creating a display layer to save as a png file to act as a folder thumbnail image - shows name of custom font file and number of characters
# Version 1.17 Adds an option to customise a line of text to include in the display information
# Version 1.18 Fixes display of smaller sized thumbnails i.e. those smaller than 401 px in height
# Version 1.19 Adds an option to remove all paths from a custom font file
# Version 1.20 .png (or extension string) can now be added to two character path names
#              accepts unicode symbol characters
# Version 1.21 Added option to delete all layers from an image:
#              This speeds up the process of adding (kerning) paths from a custom font with kerning paths to another family member without the paths since there is currently no call to open paths from a file

from gimpfu import *
from array import array
from gtk import gdk
import math
import random
import os

w = gdk.screen_width()
h = gdk.screen_height()

# Some global colours we may need to use

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkgrey = gimpcolor.RGB(50, 50, 50)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	


def do_stuff(image, layer, add_case, scale_amount, padding, old_mode, new_mode, do_remove_non_letter_layers, do_place_at_origin, do_sort, do_layer_info, 
            text_string, do_layer_names, do_path_names, do_sandwich, big_sandwich, trim, do_remove_paths, do_remove_all_letter_layers, do_display, display_credit, do_save_png):

    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    do_new_letters = False
    xcffilename = "" 
    do_merge_files = False
	
    if do_sandwich:
        			
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_names_list = image.layers
        visi_loop, top, bottom, other_layers = 0,0,0,0
		
        while visi_loop < num_layers:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "top": top += 1
            if visi_name == "bottom": bottom += 1 
            visi_loop += 1
        if top>0 or bottom>0: other_layers = 1

        
        visi_loop,first_layer = 0,-1
		
        while visi_loop < num_layers and other_layers > 0:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": 
                has_a_space = True
                space_loop = visi_loop
            if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                if first_layer == -1: first_layer = visi_loop
                # turn off visibility of letter layers
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
            # turn off visibility of letter layers
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
            visi_loop += 1
		
		
        visi_loop = 0
		
        while visi_loop < num_layers and other_layers > 0:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
                
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
            #if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                layer_width = layer_names_list[visi_loop].width
                layer_height = layer_names_list[visi_loop].height
                offset_x, offset_y = pdb.gimp_drawable_offsets(layer_names_list[visi_loop])
                if top>0:
                    # make a duplicate of the "top" layer and insert
                    top_id = pdb.gimp_image_get_layer_by_name(image,"top")
                    top_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    layer_copy = pdb.gimp_layer_copy(top_id, True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, top_position)
                    pdb.gimp_item_set_visible(layer_copy,True)
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],True) 
                    # Merge down the copied layer to the character layer
                    pdb.gimp_image_merge_down(image, layer_copy, 0)
                    active_layer = image.active_layer
                    # Re-set the id of the layer
                    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
				
                if bottom>0:
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],True) 
                    bottom_id = pdb.gimp_image_get_layer_by_name(image,"bottom")
                    bottom_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    bottom_position += 1
                    layer_copy = pdb.gimp_layer_copy(bottom_id, True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, bottom_position)
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_copy,True) 
                    layer_names_list = image.layers
                    # Merge down to the bottom layer
                    pdb.gimp_image_merge_down(image, layer_names_list[visi_loop], 0)
                    active_layer = image.active_layer
                    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
                    layer_names_list = image.layers
                    #pdb.gimp_layer_resize(layer_names_list[visi_loop], layer_width, layer_height, 0, 0)
                if not big_sandwich: pdb.gimp_layer_resize(layer_names_list[visi_loop], layer_width, layer_height,  0 - offset_x, 0 - offset_y)
				
                pdb.gimp_item_set_name(active_layer,visi_name)
                pdb.gimp_item_set_visible(active_layer,False) 
				
            visi_loop += 1
			
            
            pdb.gimp_item_set_visible(layer_names_list[first_layer],True)
            pdb.gimp_image_set_active_layer(image, layer_names_list[first_layer])

    if add_case:
        check_base_line = pdb.gimp_image_get_vectors_by_name(image,"base_line.png") #Check that a path describing the character baseline does exist
        if check_base_line>=0:
            active_layer = pdb.gimp_image_get_active_layer(image) #Remember the active layer before we begin
            char_vector = pdb.gimp_image_get_vectors_by_name(image, "base_line.png")		
            NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(char_vector)
            Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
            TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(char_vector, Char_Vector_Stroke)
            base_line_y = ListOfControlPoints[1] # Extract first control point y position
            loop = 0 # create a loop to use with each upper case letter
            while loop<=25:
                letter = chr(65 + loop) # upper case
                letter_name = letter + ".png"
                new_letter = chr(97 + loop) # lower case
                new_letter_name = new_letter + ".png"
                new_letter_name_id = pdb.gimp_image_get_layer_by_name(image,new_letter_name) # check that a lower-case letter does not already exist
                letter_name_id = pdb.gimp_image_get_layer_by_name(image,letter_name) # check that an upper-case letter exists
                if not new_letter_name_id and letter_name_id:
                    letter_name_id = pdb.gimp_image_get_layer_by_name(image,letter_name) # get layer id for existing upper case letter
                    layer_position = pdb.gimp_image_get_item_position(image, letter_name_id) # get position of this layer in the layer stack
                    layer_copy = pdb.gimp_layer_copy(letter_name_id, True) # copy the layer
                    pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)  # and insert it in the layer stack
                    original_height = layer_copy.height # get height and width of the original u/c layer
                    original_width = layer_copy.width
                    scaled_height = (original_height/100) * scale_amount # get scale factors for height and width of layer
                    scaled_width = ((original_width/100) * scale_amount)
                    layer_copy = pdb.gimp_item_transform_scale(layer_copy, 0, 0, scaled_width, scaled_height) # scale the layer to UI value
                    scaled_width = ((original_width/100) * scale_amount) + (padding * 2) # update scaling to include any new layer padding size 
                    pdb.gimp_image_select_item(image, 2, layer_copy) # get content selection from alpha
                    saved_selection = pdb.gimp_selection_save(image) # save the selection (as a channel)
                    pdb.gimp_selection_none(image) # clear the selection
                    pdb.gimp_drawable_edit_clear(layer_copy) # clear the contents of the copied layer
                    pdb.gimp_image_select_item(image, 2, saved_selection) # re-load the selection from the channel
                    success, x0, y0, x1, y1 = pdb.gimp_selection_bounds(image) # get selection information
                    translate_amount_y = base_line_y - y1 # calculate how much to move the selection in the y direction
                    pdb.gimp_selection_none(image)
                    pdb.gimp_item_transform_translate(saved_selection, 0 + padding, translate_amount_y) # move the selection
                    pdb.gimp_image_select_item(image, 2, saved_selection) # re=load the selection
                    pdb.gimp_layer_scale(layer_copy, scaled_width, original_height, False) # scale the copied layer to match height of existing layers but with new scaled width
                    pdb.gimp_drawable_edit_fill(layer_copy, 2) # fill the selection with white
                    pdb.gimp_selection_none(image)
                    pdb.gimp_item_set_name(layer_copy, new_letter_name) # re-name the new layer to lower-case custom font format
                    pdb.gimp_item_set_visible(layer_copy, False) # turn off visibility
                    pdb.gimp_image_remove_channel(image, saved_selection) # remove unwanted save selection from channel list
				
                loop += 1
            pdb.gimp_image_set_active_layer(image, active_layer) # recall original active layer

		
    if do_layer_names:
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop = 0
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8')
            if len(layer_name) == 1:
                layer_name += text_string
                pdb.gimp_item_set_name(layer_list[loop],layer_name)
            else:
                if layer_name == "space":
                    layer_name += text_string
                    pdb.gimp_item_set_name(layer_list[loop],layer_name)
            loop += 1
			
			
    if do_path_names:
        num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop < num_paths:
            path_name = pdb.gimp_item_get_name(vector_list[loop])
            path_name = path_name.decode('utf-8')
            if len(path_name) <3:
                path_name += text_string
                pdb.gimp_item_set_name(vector_list[loop],path_name)
            else:
                if path_name == "space":
                    path_name += text_string
                    pdb.gimp_item_set_name(vector_list[loop],path_name)
            loop += 1				
			
			
    if do_sort:
        space_case = [' ','space']
        upper_case = ['Z','Y','X','W','V','U','T','S','R','Q','P','O','N','M','L','K','J','I','H','G','F','E','D','C','B','A']
        lower_case = ['z','y','x','w','v','u','t','s','r','q','p','o','n','m','l','k','j','i','h','g','f','e','d','c','b','a']
        numerals = ['9','8','7','6','5','4','3','2','1','0']
        mathematical = ['-','+','=','/','*']
        punctuation = ['.',',',':',';','!','?','\'','\"']
        other_characters = ['$','%','^','&','_','{','}','@','#','~','|','\\','<','>','(',')','[',']','{','}']
        special_characters = unichr(163) + unichr(169) + unichr(174) # Pound Sterling sign, Copyright Sign, Registered Trademark sign
		
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_id_list = image.layers
        loop = 0
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_id_list[loop])
            if loop == 0:
                layer_name_list = [layer_name]
            else:
                layer_name_list.append(layer_name)
            loop += 1

        all_layers = layer_name_list
		
        for character in special_characters:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
								
        for character in other_characters:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
								
        for character in punctuation:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
						
        for character in mathematical:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in numerals:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in lower_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in upper_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
			
        for character in space_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
 
    if do_place_at_origin:
	
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
					
        loop = 0
        # Get the list of layer ids
        layer_names_list = image.layers
        
        while loop < num_layers:
            # make sure that existing layers are positioned at the origin of the image
            layer_names_list[loop].set_offsets(0, 0)
            loop += 1		

			
    if do_remove_non_letter_layers or do_remove_all_letter_layers:
        
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        do_not_remove = "legend"
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if do_remove_non_letter_layers:
                test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
                if test or "Legend" in visi_name or visi_name == "space.png" or do_not_remove.lower() in visi_name.lower():
			        # Re-position letter layers
                    layer_names_list[visi_loop].set_offsets(0, 0)
                else:
			        # Remove other layers
                    pdb.gimp_image_remove_layer(image, layer_names_list[visi_loop])
                    pdb.gimp_image_resize_to_layers(image)
            else:
			    # Remove all layers
                pdb.gimp_image_remove_layer(image, layer_names_list[visi_loop])
                
                    		
            visi_loop += 1 
			
    if trim > 0:
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
			    proc_trim_away(image, layer_names_list[visi_loop], trim)
                    		
            visi_loop += 1
			
    if do_remove_paths:
			
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop<num_vectors:
            vector_id = vector_list[loop]
            #active_vector_name = pdb.gimp_item_get_name(vector_id)
            #active_vector_name = active_vector_name.decode('utf-8')
            pdb.gimp_image_remove_vectors(image, vector_list[loop])
            loop += 1
			

    if do_layer_info:
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop = 0
        header = "This file contains the following custom font character/s:\n\n"
        layer_info = header
		
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_list[loop], layer_name)
            if test:
                if len(layer_name) == 5:
                    layer_info += layer_name[0]
                elif len(layer_name) == 6:
                    pos = layer_name.find(".png")
                    layer_info += layer_name[0:pos]
            elif (layer_name == "space.png"):
                layer_info += "space"
		
            loop += 1
        layer_info = layer_info.decode('utf-8') # Allows unicode characters above 127 to be displayed without creating errors
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(layer_info)
        pdb.gimp_message_set_handler(old_handler)
	
	
    if do_display or do_save_png:
        filename = pdb.gimp_image_get_filename(image)
        if str(filename) == "None":
            do_display, do_save_png = False, False
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("A display layer can only be created for a file that has been saved. Please save this custom font file before trying again.")
            pdb.gimp_message_set_handler(old_handler)
				
    if do_display or do_save_png:
        upper_case, lower_case, number_case, punctuation_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz", "0123456789", ".,:;'!?\""
        display_id = pdb.gimp_image_get_layer_by_name(image,"A.png") #Check for uppercase "A" - if it doesn't exist we do not carry on
        if display_id: 
            filename = pdb.gimp_image_get_filename(image) # get information about the image file name and file path
            file_extension = os.path.splitext(filename)[1]
            file_path = os.path.splitext(filename)[0]
            base_path = os.path.basename(filename)
            #pdb.gimp_message("Filename = " + str(base_path))
            visi_loop, ctr = 0,0
            num_layers, layer_ids = pdb.gimp_image_get_layers(image)
            layer_names_list = image.layers
            while visi_loop < num_layers:
                # turn off visibility of layers
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) 
                visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                visi_name = visi_name.decode('utf-8')
                test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
                if not test:
                    if visi_name.lower() <> "space.png":
                        ctr += 1	
                visi_loop += 1
				
            no_of_chars = num_layers - ctr # number of characters, including space character, contained in the layer stack
            pdb.gimp_image_set_active_layer(image, layer_names_list[0])
			
            #count the number of kerning paths in the file
            num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            loop, kern_ctr = 0,0
            while loop < num_paths:
                path_name = pdb.gimp_item_get_name(vector_list[loop])
                path_name = path_name.decode('utf-8')
                if (len(path_name) == 5 and path_name.find(".png") == 1) or (len(path_name) == 6 and path_name.find(".png") == 2):
                    kern_ctr += 1
                loop += 1				
			
			
            # Make a copy of uppercase "A" and insert it in the layer stack
            display_copy = pdb.gimp_layer_copy(display_id, True)
            pdb.gimp_image_insert_layer(image, display_copy, None, -1)
            # Rename the new layer
            pdb.gimp_item_set_name(display_copy,"Display")
            # turn on visibility of layer
            pdb.gimp_item_set_visible(display_copy, True)
			
            display_offset = display_copy.width # make a note of the width of the uppercase "A" layer
            display_id2 = pdb.gimp_image_get_layer_by_name(image,"b.png") # Make a copy of lowercase "b" if present in the stack
            if display_id2:
                display_copy2 = pdb.gimp_layer_copy(display_id2, True)
                pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
            else:
                display_id2 = pdb.gimp_image_get_layer_by_name(image,"B.png") # if lowercase "b" is not present use uppercase "B" or uppercase "A"
                if display_id2:
                    display_copy2 = pdb.gimp_layer_copy(display_id2, True)
                    pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
                else:
                    display_copy2 = pdb.gimp_layer_copy(display_id, True)
                    pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
            
            pdb.gimp_layer_set_offsets(display_copy2, display_offset, 0) # position the second display character layer
            # turn on visibility of layer
            pdb.gimp_item_set_visible(display_copy2, True)
            # merge the two characters
            display_id = pdb.gimp_image_merge_down(image, display_copy2, 0)
            # resize the image if required
            if display_id.width > image.width: pdb.gimp_image_resize(image, display_id.width, image.height, 0, 0)
            # get the text to put into the new text layer
            our_text = base_path + "\n" + str(no_of_chars) + " characters."
            if kern_ctr == 1:
                our_text += "\n" + str(kern_ctr) + " kerning path."
            else:
                our_text += "\n" + str(kern_ctr) + " kerning paths."
            if str(display_credit) <> "": our_text += "\n" + display_credit
            # add the new text layer to the layer stack
            text_layer = pdb.gimp_text_layer_new(image,our_text, "Sen", 24, 0)
            pdb.gimp_image_insert_layer(image, text_layer, None, -1)
            pdb.gimp_text_layer_set_color(text_layer, black)
            pdb.gimp_item_set_visible(text_layer, True)
            # sort out positioning for text if user text is present
            if str(display_credit) <> "": 
                if image.height < 401:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height)
                else:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height - 120)
            else:
                if image.height < 401:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height)
                else:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height - 80)
            # create a copy of the text layer to fill as a text background
            text_copy = pdb.gimp_layer_copy(text_layer, True)
            pdb.gimp_image_insert_layer(image, text_copy, None, -1)
            pdb.gimp_drawable_fill(text_layer, 2) # white fill
            # merge the text layers into the display layer
            text_copy = pdb.gimp_image_merge_down(image, text_copy, 0)
            display_id = pdb.gimp_image_merge_down(image, text_copy, 0)
            pdb.gimp_layer_set_offsets(display_id, 0, 0)
            pdb.gimp_image_resize(image, display_id.width, display_id.height, 0, 0)
			
            if do_save_png:
                # save a png file without asking the user
                save_file = file_path + ".png"
                pdb.file_png_save2(image, display_id, save_file, save_file, False, 9, True, False, False, True, True, True, True)

	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
def fn_get_mode(old_mode, new_mode):
        # Use Legacy Modes if new Default Modes are set to Normal Mode
        temp_mode = 0
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
            elif new_mode == 2:
                temp_mode = 57
            elif new_mode == 3:
                temp_mode = 59
            elif new_mode == 4:
                temp_mode = 60
            elif new_mode == 5:
                temp_mode = 36
            elif new_mode == 6:
                temp_mode = 55
            elif new_mode == 7:
                temp_mode = 31
            elif new_mode == 8:
                temp_mode = 42
            elif new_mode == 9:
                temp_mode = 33
            elif new_mode == 10:
                temp_mode = 35
            elif new_mode == 11:
                temp_mode = 54
            elif new_mode == 12:
                temp_mode = 30
            elif new_mode == 13:
                temp_mode = 43
            elif new_mode == 14:
                temp_mode = 53
            elif new_mode == 15:
                temp_mode = 23
            elif new_mode == 16:
                temp_mode = 45
            elif new_mode == 17:
                temp_mode = 44
            elif new_mode == 18:
                temp_mode = 48
            elif new_mode == 19:
                temp_mode = 49
            elif new_mode == 20:
                temp_mode = 50
            elif new_mode == 21:
                temp_mode = 51
            elif new_mode == 22:
                temp_mode = 32
            elif new_mode == 23:
                temp_mode = 52
            elif new_mode == 24:
                temp_mode = 34
            elif new_mode == 25:
                temp_mode = 46
            elif new_mode == 26:
                temp_mode = 47
            elif new_mode == 27:
                temp_mode = 41
            elif new_mode == 28:
                temp_mode = 37
            elif new_mode == 29:
                temp_mode = 38
            elif new_mode == 30:
                temp_mode = 39
            elif new_mode == 31:
                temp_mode = 40
            elif new_mode == 32:
                temp_mode = 24
            elif new_mode == 33:
                temp_mode = 25
            elif new_mode == 34:
                temp_mode = 26
            elif new_mode == 35:
                temp_mode = 27
            elif new_mode == 36:
                temp_mode = 56
            mode = temp_mode
        else:
            if old_mode == 1:
                temp_mode = 1
            elif old_mode == 2:
                temp_mode = 10
            elif old_mode == 3:
                temp_mode = 4
            elif old_mode == 4:
                temp_mode = 16
            elif old_mode == 5:
                temp_mode = 7
            elif old_mode == 6:
                temp_mode = 9
            elif old_mode == 7:
                temp_mode = 3
            elif old_mode == 8:
                temp_mode = 17
            elif old_mode == 9:
                temp_mode = 5
            elif old_mode == 10:
                temp_mode = 19
            elif old_mode == 11:
                temp_mode = 18
            elif old_mode == 12:
                temp_mode = 6
            elif old_mode == 13:
                temp_mode = 8
            elif old_mode == 14:
                temp_mode = 20
            elif old_mode == 15:
                temp_mode = 21
            elif old_mode == 16:
                temp_mode = 22
            elif old_mode == 17:
                temp_mode = 11
            elif old_mode == 18:
                temp_mode = 12
            elif old_mode == 19:
                temp_mode = 13
            elif old_mode == 20:
                temp_mode = 14
            else:
                temp_mode = 0
            mode = temp_mode
			

        return mode
	
def proc_trim_away(img, layer, amount):
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################


register(
    "python_fu_Custom_Font_Layer_Utils_v1-21",
    "Provides options for custom font layers",
    "Provides options for custom font layers",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2019",
    "<Image>/Filters/Custom Font Tools/Custom Font Layer Utils v1.21...",
    "*",
    [
        (PF_TOGGLE, "add_case", "Create lower-case \nfrom upper-case:", 0),
        (PF_SPINNER, "scale_letter", "Scale Character Amount (%): ", 75,(25,95,1)),
        (PF_SPINNER, "pad_scaling", "Padding size for \nscaled letters (px): ", 0,(0,100,1)),
        (PF_OPTION, "mode_option", "Legacy Mode \nto use with font face:", 0, ("Normal", "Dissolve", "Lighten", "Screen", "Dodge", "Addition", "Darken", "Multiply", "Burn", "Overlay", "Softlight", "Hardlight", "Difference", "Subtract", "Grain Extract", "Grain Merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSV)")),
        (PF_OPTION, "mode_option2", "New Mode \nto use with font face:", 0, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance")),
        (PF_TOGGLE, "remove", "Remove all non-font layers", 0),
        (PF_TOGGLE, "origin_option", "Ensure that all layers \nare positioned to the origin:", 0),
        (PF_TOGGLE, "sort_option", "Sort layers into order:", 0),
        (PF_TOGGLE, "info_option", "Display current characters \nin a message box:", 0),
        (PF_STRING, "word", "Label String:", ".png"),
        (PF_TOGGLE, "layer_option", "Add Label String to \nsingle letter layer labels:", 0),
        (PF_TOGGLE, "path_option", "Add Label String to single\nor double letter path labels:", 0),
        (PF_TOGGLE, "sandwich_option", "Sandwich characters between \nlayers called top and bottom:", 0),
        (PF_TOGGLE, "restrict_option", "Set sandwich layer \nto canvas size:", 0),
        (PF_SPINNER, "pf_trim", "Trim pixels from character: ", 0,(0,10,1)),
        (PF_TOGGLE, "remove_paths", "Remove all vectors from the image:", 0),
        (PF_TOGGLE, "remove_all_layers", "Remove ALL LAYERS from the image:", 0),
        (PF_TOGGLE, "dopng", "Add a display layer:", 0),
        (PF_STRING, "pf_credit", "Display credit: ", "Decoration by Skinnyhouse."),
        (PF_TOGGLE, "save_png", "Save display layer as png image:", 0),
        #(PF_DIRNAME, "p28", "Directory to use:", 0)
    ],
    [],
    do_stuff)

main()

	

