#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2019 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Lighting Effects vX.XX...'
#
# This plug-in applies the lighting plug-in to the individual character layers of a custom font file
# The basic work-flow: Take a character layer>make a copy>blur the copy>add a white layer below>merge the blurred layer with the background layer to create a bumpmap layer
#                      > use the first available non-character layer and apply the lighting plug-in
# 
# The plug-in was inspired by some of the custom fonts created by Alan and uploaded to GimpLearn in 2019
# It broadly follows the technique developed in 2013 by Gnu Tux, and added to by others, to produce environment mapped text. The tutorial can be viewed here: http://gimpchat.com/viewtopic.php?f=23&t=8423&sid=97475303608c48de61d059671d94ba49
# The Lighting plug-in does not work as it should in non-interactive mode - when called within the body of another plug-in - as Graechan bemoaned on Gimpchat in this thread: http://gimpchat.com/viewtopic.php?f=8&t=2895&hilit=Saulgoode+lighting&sid=1134b9209fe81b15b8da3890a0042057
# In particular I have struggled to use the environment layer and the diffuse reflectivity parameters in non-interactive mode in the lighting plug-in
# Thanks to the detective work of Graechan, Saulgoode and Gnu Tux I have at least been able to devise a means of working around these difficulties to produce some interesting effects

# Change Log
# Version 0.00 bones of plug-in with lighting parameters set from dialog UI
# Version 0.01 added more options:
#              replace light direction or position with compass points
#              range of greys used to create bumpmap for basic characters
#              to invert character bumpmap
#              control depth of bumpmap applied to image layer
#              order of blur
#              distort image applied to character with apply lens and/or polar coords plug-ins
#              mode selection for applying image layer
#              selective blur for minimizing artifacts
#              re-apply lighting
# Version 0.02 Added some gradient fill options when creating bumpmap layer, 
#              Added bigger range of greys to use as background when creating bumpmap layer, 
#              Altered code for re-applying lighting - now creates a new layer to merge with existing layers
#              Updated legend info
# Version 0.03 Re-ordered overlay so that a new character is applied on top of its old character as I originally intended - a Homer Simpson moment
#              Added an option to allow the user to choose whether to apply the original character outline - or not
# Version 0.04 Removed light source option - now uses directional light only (no observable difference between the two options in any case)
#              Added option to allow random positioning when copying and pasting from environment layer onto character layers
# Version 0.05 Accepts unicode symbol characters

from gimpfu import *
import random
from array import array

def do_stuff(image, layer, fg_light, compass, light_amount, light_z, 
    intensity, light_glow, light_shiny, light_polished,
    bump_grey, bg_bump_colour, use_more_bump, do_random, order_of_blur, bumpmap_depth, bump_blur, bump_fill_type, 
    use_lens, refraction, use_polar, old_mode, new_mode, blur_radius, blur_delta, do_overlay, overlay_opacity, retain_outline):

    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    #use_clipboard_as_brush, do_layer_info, do_layer_names, do_path_names, do_sort = 0,0,0,0,0
    outline_colour = pdb.gimp_context_get_foreground()
	
    # Define some colours
    black = gimpcolor.RGB(0, 0, 0)
    white = gimpcolor.RGB(255, 255, 255)
    light_grey = gimpcolor.RGB(220, 220, 220)
    midgrey = gimpcolor.RGB(125, 125, 125)
    darkgrey = gimpcolor.RGB(42, 42, 42)
	
    bump_col, grey_info = fn_get_grey(bump_grey)
    bg_bump_col,bg_grey_info = fn_get_bg_bump_col(bg_bump_colour)
	
    use_lighting = True
    mode, n_mode_info, l_mode_info = fn_get_mode(old_mode, new_mode)
	
    light_bright = 0.5
	
    if do_overlay:
        type_info = "overlaid"
    else:
        type_info = "new"
		
    light_source = 1 # Hard fix type of light used to directional
    
    if light_source == 1:
        light_source_info = "directional"
    else:
        light_source_info = "point"
		
		
    if use_more_bump:
        face_bump_info = "Yes"
    else:
        face_bump_info = "No"

    invert_bump = 1	
    if invert_bump:
        invert_info = "Yes"
    else:
        invert_info = "No"
		
    if use_lens:
        lens_info = "Yes"
    else:
        lens_info = "No"
	
    if order_of_blur == 0:
        #blur_before_bump = True
        order_info = "Bumpmap after blurring"
    elif order_of_blur == 1:
        #blur_before_bump = False
        order_info = "Bumpmap before blurring"
    else:
        #blur_before_bump = False
        order_info = "No blur used"
		
    if use_polar:
        polar_info = "Yes"
    else:
        polar_info = "No"
	
    #use current foreground colour as light or white light
    if fg_light:
        light_colour = fn_set_lighting_colour(pdb.gimp_context_get_foreground())
    else:
        light_colour = (255,255,255) #white light
		
    # Set lighting plug-in pos and dir xyz values from compass point selected via dialog UI
    if compass == 0:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 0,0,-1,-1 #S
        effect_azimuth = 90
        compass_info = "N"
    elif compass == 1:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,-1,-1 #SW
        effect_azimuth = 45
        compass_info = "NE"
    elif compass == 2:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,0,0 #W
        effect_azimuth = 0
        compass_info = "E"
    elif compass == 3:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,1,1 #NW
        effect_azimuth = 315
        compass_info = "SE"
    elif compass == 4:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 0,0,1,1 #N
        effect_azimuth = 270
        compass_info = "S"
    elif compass == 5:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,1,1 #NE
        effect_azimuth = 225
        compass_info = "SW"
    elif compass == 6:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,0,0 #E
        effect_azimuth = 180
        compass_info = "W"
    else:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,-1,-1 #SE
        effect_azimuth = 135
        compass_info = "NW"
		
    # set non UI bumpmap plugin parameters here to reduce height of dialog
    effect_elevation = 35 # possible temp elevation
    #bumpmap_depth = 10 # possible temp bumpmap depth
    waterlevel = 0
    ambient = 0
    
    light_pos_x *= light_amount
    light_pos_y *= light_amount
    light_dir_x *= light_amount
    light_dir_y *= light_amount
	
	
    light_pos_z, light_dir_z = light_z,light_z
	
    # set non UI shine plugin parameters
    shine_shadow_X, shine_shadow_Y, shine_shadow_size, shine_shadow_opacity, shine_keep_selection, shine_keep_layers = 0,0,0,0,0,0
    loop_ptr = 1
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
	
    visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
    bumpmap_layer, environment_layer = "none", "none"
	
    if use_lighting:
        while visi_loop < num_layers:
            # check for layers to use as bumpmap source
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if len(visi_name) == 1 or visi_name == "space.png" or len(visi_name) == 5 and ".png" in visi_name:
                if first_letter_layer < 0: first_letter_layer = visi_loop
            else:
                pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
                is_empty = pdb.gimp_selection_is_empty(image)
                if is_empty == False: 
                    # If the lighting plug-in worked correctly then the following 3 lines would be needed
                    # Test for an alpha layer
                    #has_alpha = pdb.gimp_drawable_has_alpha(layer_names_list[visi_loop])
                    #if (not has_alpha): other_layers += 1
                    other_layers += 1
                    if other_layers == 1: 
                        # Use the top-most non-character layer as the source for the environment map layer
                        environment_layer = layer_names_list[visi_loop]
                        environment_layer_info = pdb.gimp_item_get_name(layer_names_list[visi_loop]) # more info for Legend layer
                        #environment_layer = layer_ids[visi_loop]
                        #pdb.gimp_message("There is a bump layer without an alpha channel here. : " + str(layer_names_list[visi_loop]) )
                        # Use the top-most non-character layer as the source for the bumpmap layer and make a copy of it
                        #bumpmap_layer = layer_names_list[visi_loop].copy()
                        #image.add_layer(bumpmap_layer, -1)
                        #bumpmap_layer_name = "Bump"
                        #pdb.gimp_image_set_active_layer(image, bumpmap_layer)
                        pdb.gimp_selection_none(image)
                        # desaturate the newly created bumpmap layer
                        #pdb.gimp_drawable_desaturate( bumpmap_layer, 3)
                    
                    pdb.gimp_selection_none(image)	
				
            visi_loop += 1
        # Update the lists of layers
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
    if other_layers == 0: 
        use_lighting = False # if there are no layers to use for the bumpmap source then cancel it
        if use_lighting == False:
            message_text = "There appears to be no suitable non-character layer to use."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)
    else:
	
        
        if use_lens or use_polar:
            environment_copy = pdb.gimp_layer_copy(environment_layer, True)
            pdb.gimp_image_insert_layer(image, environment_copy, None, 0)
            pdb.gimp_item_set_visible(environment_copy, True)
            e_width = environment_copy.width
            e_height = environment_copy.height
            if e_width>image.width:
			    offx = 0 - ((e_width - image.width)/2)
            else:
                offx = 0
            if e_height>image.height:
                offy = 0-((e_height - image.height)/2)
            else:
                offy = 0
            pdb.gimp_layer_set_offsets(environment_copy,offx,offy)

        if use_lens and not use_polar:
            pdb.plug_in_applylens(image,environment_copy,refraction,True,False,False)
            environment_layer = environment_copy
    
        if use_polar and not use_lens:
            pdb.plug_in_polar_coords(image,environment_copy,100,180,False,True,False)
            environment_layer = environment_copy
			
        if use_lens and use_polar:
            pdb.plug_in_polar_coords(image,environment_copy,100,180,False,True,False)
            pdb.plug_in_applylens(image,environment_copy,refraction,True,False,False)
            environment_layer = environment_copy
		
		
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
			
		
	
    first_char_layer_name = "none"
	
    if not ( use_lighting == False):
        # turn off visibility of layers
        visi_loop, first_char_layer, legend_ctr  = 0, -1, 0
        while visi_loop < num_layers:
	        # get info about layer we are working on
            layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            layer_name = layer_name.decode('utf-8')
            legend_num, legend_name = 0, "None"
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], layer_name)
            if test:
            #if len(layer_name) == 1 or len(layer_name) == 5 and ".png" in layer_name:
                if first_char_layer < 0: 
                    first_char_layer = visi_loop
                    first_char_layer_name = layer_name
            else:
                if "Legend" in layer_name:# check for a legend layer that might contain coordinate information
                    legend_ctr += 1
                    if layer_name == "Legend": 
                        legend_name = "Legend"
                    if "#" in layer_name:
                        new_legend_num = max(layer_name)
                        if new_legend_num > legend_num:legend_name = layer_name
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)
            visi_loop += 1
        if legend_ctr == 0: # No legend layer so cancel overlay
            do_overlay = False
            type_info = "new"
				
		
			
        if do_overlay:
            if do_random:
                # Look for any coordinate information contained in the first part of legend layer
                legend_text_layer = pdb.gimp_image_get_layer_by_name(image, legend_name)
                legend_text = pdb.gimp_text_layer_get_text(legend_text_layer)
                new_text = legend_text.splitlines()
                legend_text = new_text[0] 
                random_list = legend_text
			
                index = random_list.find(",")
                iloop, icounter = index, 0
                while index > -1:
                    index = legend_text.find(",", iloop + 1)
                    if index > -1: 
                        temp = int(legend_text[iloop + 3:index - 1])
                        if icounter == 0:
                            new_list = [temp]
                            icounter += 1
                        else:
                            new_list.append(temp)
                        iloop = index
                temp = int(legend_text[iloop + 3:len(legend_text)-2]) 
                new_list.append(temp)
                random_list = new_list
                lp = range(0,len(random_list),3)
                for iloop in lp:
                    temp = random_list[iloop]
                    if iloop == 0:
                        name_random = [temp]
                    else:
                        name_random.append(temp)
            else:
                # No coordinates required for a fixed position so use the name of the environment layer here
                random_list = [str(environment_layer)]
                name_random = random_list
			
        else:
            # No coordinates required for a fixed position so use the name of the environment layer here
            random_list = [str(environment_layer)]
            name_random = random_list
			
        while loop >= 0:
	        # get info about layer we are working on
            layer_id = layer_ids[loop]
            layer_name = layer_names_list[loop]
	
            # cater for unicode characters above normal range
            display_layer_name =  pdb.gimp_item_get_name(layer_names_list[loop])
            display_layer_name = display_layer_name.decode('utf-8')	
	
            # check we have the layer id of the currently active layer
            pdb.gimp_image_set_active_layer(image, layer_name)
            layer_id = pdb.gimp_image_get_active_layer(image)

            # check to see if the current layer is a custom font character layer
            test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
				
            if test_for_custom_layer:

                # ensure we have the active layer and make it visible
                active_layer_id = pdb.gimp_image_get_active_layer(image)
                pdb.gimp_item_set_visible(active_layer_id, True)
  
                if use_lighting:
								
                    active_layer = image.active_layer 
					 	
                    if do_overlay: 
                        # make a copy of the original character so that we can use it later to meld with the newly applied lighting effects
                        copy_of_original = pdb.gimp_layer_copy(active_layer, True)
                        image.add_layer(copy_of_original, -1)
                        pdb.gimp_item_set_visible(copy_of_original, False)
                        pdb.gimp_item_set_name(copy_of_original,"Copy of Original")
                        pdb.gimp_image_set_active_layer(image, layer_name)
                        layer_id = pdb.gimp_image_get_active_layer(image)
                        active_layer = image.active_layer 
						
		
                    # save a copy of the character's selection to use later for tidying up the finished character
                    # alpha select current character
                    pdb.gimp_image_select_item(image, 2, active_layer)
                    #save a copy of this selection
                    saved_selection = pdb.gimp_selection_save(image)
                    # deselect
                    pdb.gimp_selection_none(image)
                    # call proc_do_environment to do the actual work
                    active_layer_id = proc_do_environment(image, active_layer, bump_col, bg_bump_col, bump_blur, bump_fill_type, environment_layer, 
                    light_colour, light_source, 
                    light_pos_x, light_pos_y, light_pos_z, light_dir_x, light_dir_y, light_dir_z, 
                    light_glow, intensity, light_bright, light_shiny, light_polished, 
                    blur_radius, blur_delta, bumpmap_depth, waterlevel, ambient, 
                    effect_elevation, effect_azimuth, use_more_bump, order_of_blur, do_overlay, do_random, display_layer_name, random_list, name_random)	
					                    					
                    # make a copy of the environment layer to be overlaid onto the character with the mode selected from the dialog UI
                    layer_copy = pdb.gimp_layer_copy(environment_layer, True)
                    # insert the layer and make it visible
                    pdb.gimp_image_insert_layer(image, layer_copy, None, -1)
                    pdb.gimp_item_set_visible(layer_copy, True)
                    #if not do_random: pdb.gimp_layer_set_offsets(layer_copy,0,0)
                    if do_random: pdb.gimp_layer_set_offsets(layer_copy,0,0)
                    # set the layer mode
                    pdb.gimp_layer_set_mode(layer_copy, mode)
					
                    # merge the layers
                    active_layer = pdb.gimp_image_merge_down(image, layer_copy,2)
                    # reset the active layer
                    active_layer_id = pdb.gimp_image_get_active_layer(image)
                    # reposition character layer to image origin
                    if not do_random: pdb.gimp_layer_set_offsets(active_layer_id,0,0)
                    if do_random: pdb.gimp_layer_set_offsets(active_layer_id,0,0)
					
                    if do_overlay:
                        # use the copy of the original character and meld it with the new character according to the user set mode
                        # make the layer visible
                        pdb.gimp_item_set_visible(copy_of_original, True)
                        # lower the copy of the original character below the new character
                        pdb.gimp_image_lower_item(image, copy_of_original)
                        # set the layer mode for the new version of the character
                        pdb.gimp_layer_set_mode(active_layer, mode)
                        # set the layer mode opacity as per user setting
                        pdb.gimp_layer_set_opacity(active_layer, overlay_opacity)
                        # merge the two versions of the character
                        active_layer = pdb.gimp_image_merge_down(image, active_layer,2)
                        # rename the merged layer with the character name
                        pdb.gimp_item_set_name(active_layer, str(display_layer_name))
                        # reset the active layer
                        active_layer_id = pdb.gimp_image_get_active_layer(image)
						
                    if retain_outline:
                        # load character selection from saved channel
                        pdb.gimp_image_select_item(image, 2, saved_selection)
				        # shrink selection
                        pdb.gimp_selection_shrink(image,1)
                        # invert it
                        pdb.gimp_selection_invert(image)
                        # clear pixels 
                        success = pdb.gimp_edit_cut(active_layer)
                        # deselect
                        pdb.gimp_selection_none(image)
                    # remove the character selection from the image
                    pdb.gimp_image_remove_channel(image, saved_selection)
				
                if (num_layers - loop_ptr) > 0: pdb.gimp_item_set_visible(active_layer_id, False)
			
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
			
            
			
            loop_ptr += 1
            loop = num_layers - loop_ptr
   

		
        if (use_lens or use_polar): pdb.gimp_image_remove_layer(image,environment_layer)
	
    if use_lighting:
        random_list[0] = environment_layer_info
        add_legend = True # temp hard value
        if add_legend:
            text_legend = str(random_list)
            text_legend += "\nEnvironment layer = " + environment_layer_info
            text_legend += "\nLight Colour = rgb: " + str(light_colour[0]) + "," + str(light_colour[1]) + "," + str(light_colour[2]) + "\nLight Type = " + light_source_info + "\nLight Direction = " + compass_info
            text_legend += "\nDistance from light source = " + str(light_amount) + "\nElevation of light source = " + str(light_z) + "\nLight intensity = " + str(intensity) + "\nGlow = " + str(light_glow)
            text_legend += "\nBrightness = " + str(light_bright) + "\nShiny = " + str(light_shiny) + "\nPolished = " + str(light_polished)
            text_legend += "\nBumpmap characters = " + face_bump_info + "\nBlur order = " + order_info + "\nBumpmap depth = " + str(bumpmap_depth) + "\nAmount of blur = " + str(bump_blur)
            text_legend += "\nBumpmap FG Grey = " + grey_info + "\nBumpmap BG Grey = " + bg_grey_info + "\nApply Lens = " + lens_info + "\nLens index = " + str(refraction) + "\nApply polar coords = " + polar_info + "\nLegacy mode = " + l_mode_info + "\nNew mode = " + n_mode_info + "\nArtifact radius = " + str(blur_radius) + "\nArtifact delta = " + str(blur_delta) + "\nType of characters = " + type_info
            if do_overlay: text_legend += "\nOpacity of overlaid characters = " + str(overlay_opacity)
            proc_add_legend(image, text_legend, "Sen", black, 24)
			
    if first_char_layer_name <> "none":
        first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
        pdb.gimp_item_set_visible(first_char_layer_id, True)
        pdb.gimp_image_set_active_layer(image,first_char_layer_id)
		
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
		
#### End of do_stuff

######################################################################################################################################

def proc_add_legend(image, text_legend, font_name, text_colour, text_size):
    legend_layer = pdb.gimp_text_layer_new(image,text_legend, font_name, text_size, 0)
    pdb.gimp_image_insert_layer(image, legend_layer, None, -1)
    pdb.gimp_text_layer_set_color(legend_layer, text_colour)
    pdb.gimp_item_set_visible(legend_layer, False)
    pdb.gimp_image_lower_item_to_bottom(image, legend_layer)
    pdb.gimp_item_set_name(legend_layer,"Legend")

######################################################################################################################################

def proc_do_environment(image, active_layer, colour_fg, colour_bg, bump_blur, bump_fill_type, environment_layer, light_colour, 
    light_source, light_pos_x, light_pos_y, light_pos_z, light_dir_x, light_dir_y, light_dir_z, light_glow, intensity, light_bright, light_shiny, light_polished, 
    blur_radius, blur_delta, bumpmap_depth, waterlevel, ambient, effect_elevation, effect_azimuth, use_more_bump, order_of_blur, do_overlay, do_random, display_layer_name, random_list, name_random):

		
    work_layer = active_layer
    work_layer_name = work_layer.name
    width = work_layer.width
    height = work_layer.height
	
    random_x, random_y = 0,0
    if do_random:
                      
        random.seed()
        random_x = random.randrange(0, (environment_layer.width + 1) - work_layer.width)
        random.seed()
        random_y = random.randrange(0, (environment_layer.height + 1) - work_layer.height)
        
        if do_overlay:
            # Look for coordinate information
            temp = ord(display_layer_name[0])
            if temp in name_random:
                index = name_random.index(temp) # pointer to current character layer
                random_x = int(random_list[index + 1] ) # pointer to x coordinate
                random_y = int(random_list[index + 2] ) # pointer to y coordinate
            else:
                random_x = 0
                random_y = 0
			
        else:
            # add character and coordinates to the list of information for random positions used on environment layer
            random_list.append(str(ord(display_layer_name[0])))
            random_list.append(str(random_x))
            random_list.append(str(random_y))
        
	 
    
    # alpha select current character
    pdb.gimp_image_select_item(image, 2, work_layer)
    # enlarge the selection by 1 pixel
    pdb.gimp_selection_grow(image, 1)
    success = pdb.gimp_edit_cut(work_layer)
    # shrink the selection by 1 pixel
    pdb.gimp_selection_shrink(image, 1)
    #set fg colour
    pdb.gimp_context_set_foreground(colour_fg)
    #set bg colour
    pdb.gimp_context_set_background(colour_bg)
    if bump_fill_type == 0:
        # fill selection
        pdb.gimp_drawable_edit_fill(work_layer,0)
    else:
        pdb.gimp_context_set_gradient_fg_bg_rgb()
        gradient_x_start = width/2
        gradient_y_start = 0 + (height/2)
        gradient_x_end = width
        gradient_y_end = 0 + height
        if bump_fill_type == 1: #Linear Gradient
            gradient_x_start = 0
            gradient_y_start = 0
            pdb.gimp_drawable_edit_gradient_fill(work_layer, 0,0,False,1,1,True, gradient_x_start, gradient_y_start, gradient_x_end, gradient_y_end)
        elif bump_fill_type == 2: #Radial Gradient
            pdb.gimp_drawable_edit_gradient_fill(work_layer, 2,0,False,1,1,True, gradient_x_start, gradient_y_start, gradient_x_end, gradient_y_end)
        elif bump_fill_type == 3: #Square Gradient
            pdb.gimp_drawable_edit_gradient_fill(work_layer, 3,0,False,1,1,True, gradient_x_start, gradient_y_start, gradient_x_end, gradient_y_end)
        elif bump_fill_type == 4: #Conical Symmetric Gradient
            pdb.gimp_drawable_edit_gradient_fill(work_layer, 4,0,False,1,1,True, gradient_x_start, gradient_y_start, gradient_x_end, gradient_y_end)
        elif bump_fill_type == 5: #Conical Asymmetric Gradient
            pdb.gimp_drawable_edit_gradient_fill(work_layer, 5,0,False,1,1,True, gradient_x_start, gradient_y_start, gradient_x_end, gradient_y_end)
			
    # deselect
    pdb.gimp_selection_none(image)
    copy_layer = pdb.gimp_layer_copy(work_layer, True)
    image.add_layer(copy_layer, -1)
    pdb.gimp_image_lower_item(image, copy_layer)
		
    bump_layer = gimp.Layer(image, "Bump Layer", width, height, RGBA_IMAGE, 100, 0)
    image.add_layer(bump_layer, -1)
    pdb.gimp_image_lower_item(image, bump_layer)
    bump_layer.fill(1)
    pdb.gimp_image_merge_down(image, copy_layer,0)
    bump_layer = image.active_layer 	
    
    pdb.gimp_layer_set_offsets(bump_layer, random_x, random_y)
	
    if order_of_blur == 0:
        # Blur - bumpmap - blur
        pdb.plug_in_gauss_iir(image, bump_layer, bump_blur/2, True, True)
        if use_more_bump: bump_layer = proc_do_lighting_bump(image, environment_layer, bump_layer, bumpmap_depth, waterlevel, ambient, effect_elevation, effect_azimuth, random_x, random_y)
        pdb.plug_in_gauss_iir(image, bump_layer, bump_blur/2, True, True)
    elif order_of_blur == 1:
        # Bumpmap - blur
        if use_more_bump: bump_layer = proc_do_lighting_bump(image, environment_layer, bump_layer, bumpmap_depth, waterlevel, ambient, effect_elevation, effect_azimuth, random_x, random_y)	
        pdb.plug_in_gauss_iir(image, bump_layer, bump_blur, True, True)	
    else:
        # Blur - bumpmap
        pdb.plug_in_gauss_iir(image, bump_layer, bump_blur, True, True)	
        if use_more_bump: bump_layer = proc_do_lighting_bump(image, environment_layer, bump_layer, bumpmap_depth, waterlevel, ambient, effect_elevation, effect_azimuth, random_x, random_y)
            
    
	
  	# Set variables for proc_lighting

    bump_map = bump_layer

    env_map = environment_layer
    #pdb.gimp_message("Environment Layer =  " + str(env_map))
    bump_map_flag = True
    env_map_flag = True
    bump_map_curve = 2 # sinusuidol
    #light_source = 1
    #light_colour = (255,255,255)
    #light_colour = (251,212,8)
    #pdb.gimp_message("light_colour =  " + str(light_colour) + " : temp_light_colour = " + str(temp_light_colour))
    #light_pos_x = -1.00
    #light_pos_y = -1.00
    #light_pos_z = 1.00
    #light_dir_x = -1.50
    #light_dir_y = -1.50
    #light_dir_z = 2.50
    #light_glow = 0.60 # 0.60
    #light_bright =  0.50 # 0.50
    #light_diffuse_reflec = 1.00 #2.80
    #light_polished = 20 #27
    #light_shiny = 1.00 # 1.00
    anti_aliasing = True
    new_light_image = False
    transparent_bg = False
	
	
    pdb.gimp_image_set_active_layer(image, work_layer)
	
    if do_random:
        pdb.gimp_layer_set_offsets(work_layer, random_x, random_y)
	
    else:
        # centre character layer within image  before using lighting plug-in
        if work_layer.width < (image.width - 2):
            offx = (image.width - work_layer.width)/2
            if offx<0:offx = 0
        else:
            offx = 0
        if work_layer.height < (image.height - 2):
            offy = 0 - (image.height - work_layer.height)/2
            if offy<0:offy = 0
        else:
            offy = 0
        pdb.gimp_layer_set_offsets(work_layer,offx,offy)
	

    #REM Step 11: Call Lighting plugin
    proc_lighting(image, work_layer, bump_map, env_map, bump_map_flag, env_map_flag, bump_map_curve, light_source, light_colour, 
    light_pos_x, light_pos_y, light_pos_z,
    light_dir_x, light_dir_y, light_dir_z, 
    light_glow, intensity, light_bright, light_shiny, light_polished, 
    anti_aliasing, new_light_image, transparent_bg)
	
    #if not do_overlay: image.remove_layer(bump_layer)
    image.remove_layer(bump_layer)
	
    pdb.gimp_image_set_active_layer(image, work_layer)
    active_layer = image.active_layer
	
    # reposition character layer to image origin
	##### Try commenting out
    #if not do_random: pdb.gimp_layer_set_offsets(work_layer,0,0)
    
    pdb.plug_in_sel_gauss(image, active_layer, blur_radius, blur_delta)	
	
    active_layer = image.active_layer 
    return active_layer
	
	
######################################################################################################################################

def fn_set_lighting_colour(colour):
    red = colour[0]
    green = colour[1]
    blue = colour[2]
    new_colour = (int(red), int(green),int(blue))
    return new_colour

######################################################################################################################################

def fn_get_grey(grey):
    if grey == 0:
        grey_shade = gimpcolor.RGB(0, 0, 0)
        grey_info = "Black"
    elif grey == 1:
        grey_shade = gimpcolor.RGB(44, 44, 44)
        grey_info = "Grey 1"
    elif grey == 2:
        grey_shade = gimpcolor.RGB(60, 60, 60)
        grey_info = "Grey 2"
    elif grey == 3:
        grey_shade = gimpcolor.RGB(76, 76, 76)
        grey_info = "Grey 3"
    elif grey == 4:
        grey_shade = gimpcolor.RGB(92, 92, 92)
        grey_info = "Grey 4"
    elif grey == 5:
        grey_shade = gimpcolor.RGB(108, 108, 108)
        grey_info = "Grey 5"
    elif grey == 6:
        grey_shade = gimpcolor.RGB(124, 124, 124)
        grey_info = "Grey 6"
    elif grey == 7:
        grey_shade = gimpcolor.RGB(140, 140, 140)
        grey_info = "Grey 7"
    elif grey == 8:
        grey_shade = gimpcolor.RGB(156, 156, 156)
        grey_info = "Grey 8"
    elif grey == 9:
        grey_shade = gimpcolor.RGB(172, 172, 172)
        grey_info = "Grey 9"
    elif grey == 10:
        grey_shade = gimpcolor.RGB(188, 188, 188)
        grey_info = "Grey 10"
    elif grey == 11:
        grey_shade = gimpcolor.RGB(204, 204, 204)
        grey_info = "Grey 11"
    elif grey == 12:
        grey_shade = gimpcolor.RGB(220, 220, 220)
        grey_info = "Grey 12"
    elif grey == 13:
        grey_shade = gimpcolor.RGB(236, 236, 236)
        grey_info = "Grey 13"
    elif grey == 14:
        grey_shade = gimpcolor.RGB(242, 242, 242)
        grey_info = "Grey 14"
    else:
        grey_shade = gimpcolor.RGB(255, 255, 255)
        grey_info = "White"

    return grey_shade, grey_info

######################################################################################################################################


def fn_get_bg_bump_col(grey):
    if grey == 0:
        grey_shade = gimpcolor.RGB(255,255,255)
        grey_info = "White"
    elif grey == 1:
        grey_shade = gimpcolor.RGB(242, 242, 242)
        grey_info = "Grey 14"
    elif grey == 2:
        grey_shade = gimpcolor.RGB(236, 236, 236)
        grey_info = "Grey 13"
    elif grey == 3:
        grey_shade = gimpcolor.RGB(220, 220, 220)
        grey_info = "Grey 12"
    elif grey == 4:
        grey_shade = gimpcolor.RGB(204, 204, 204)
        grey_info = "Grey 11"
    elif grey == 5:
        grey_shade = gimpcolor.RGB(188, 188, 188)
        grey_info = "Grey 10"
    elif grey == 6:
        grey_shade = gimpcolor.RGB(172, 172, 172)
        grey_info = "Grey 9"
    elif grey == 7:
        grey_shade = gimpcolor.RGB(156, 156, 156)
        grey_info = "Grey 8"
    elif grey == 8:
        grey_shade = gimpcolor.RGB(140, 140, 140)
        grey_info = "Grey 7"
    elif grey == 9:
        grey_shade = gimpcolor.RGB(124, 124, 124)
        grey_info = "Grey 6"
    elif grey == 10:
        grey_shade = gimpcolor.RGB(108, 108, 108)
        grey_info = "Grey 5"
    elif grey == 11:
        grey_shade = gimpcolor.RGB(92, 92, 92)
        grey_info = "Grey 4"
    elif grey == 12:
        grey_shade = gimpcolor.RGB(76, 76, 76)
        grey_info = "Grey 3"
    elif grey == 13:
        grey_shade = gimpcolor.RGB(60, 60, 60)
        grey_info = "Grey 2"
    elif grey == 14:
        grey_shade = gimpcolor.RGB(44, 44, 44)
        grey_info = "Grey 1"
    else:
        grey_shade = gimpcolor.RGB(0, 0, 0)
        grey_info = "Black"

    return grey_shade,grey_info

######################################################################################################################################

def proc_do_lighting_bump(image, bumpmap_layer_id, active_layer, bumpmap_depth, waterlevel, ambient, effect_elevation, effect_azimuth, random_x, random_y):
    work_layer = active_layer
    work_layer_name = work_layer.name
    width = work_layer.width
    height = work_layer.height
    bumpmap_depth = bumpmap_depth * 0.65
    if bumpmap_depth < 1: bumpmap_depth = 1
		
    if waterlevel > 0:
        waterlevel = waterlevel/100
        waterlevel = 1.0 - waterlevel
    else:
        waterlevel = 0.0
			
    if ambient > 0:
        ambient = ambient/100
        ambient = 1.0 - ambient
    else:
        ambient = 0.0        

    pdb.plug_in_bump_map(image, work_layer, bumpmap_layer_id, effect_azimuth, effect_elevation, bumpmap_depth, random_x, random_y, waterlevel, ambient, True, False, 0)
		
    active_layer = image.active_layer 
    return active_layer
        
######################################################################################################################################

	
def proc_lighting(effect_image, effect_drawable, bump_drawable, env_drawable, do_bump, do_env, 
    bump_type, light_type, light_colour, 
    light_pos_x, light_pos_y, light_pos_z, light_dir_x, light_dir_y, light_dir_z, 
    glow_ambient_intensity, intensity_diffuse_intensity, bright_diffuse_reflectivity, 
    shiny_spec_reflectivity, polish_material_highlight, 
    anti_aliasing, new_image, transparent_bg):
    
    # Rem: Use the lighting Plugin
    #LIGHT 1 

    pdb.plug_in_lighting(effect_image, effect_drawable, bump_drawable, env_drawable, do_bump, do_env, 
    bump_type, light_type, light_colour, 
    light_pos_x, light_pos_y, light_pos_z, light_dir_x, light_dir_y, light_dir_z, 
    glow_ambient_intensity, intensity_diffuse_intensity, bright_diffuse_reflectivity, 
    shiny_spec_reflectivity, polish_material_highlight, 
    anti_aliasing, new_image, transparent_bg)

        
######################################################################################################################################

	
def fn_get_mode(old_mode, new_mode):
        # Use Legacy Modes if new Default Modes are set to Normal Mode
        temp_mode = 0
        l_mode_info, n_mode_info = "Normal", "Normal"
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
                n_mode_info = "Dissolve"
            elif new_mode == 2:
                temp_mode = 57
                n_mode_info = "Color Erase"
            elif new_mode == 3:
                temp_mode = 59
                n_mode_info = "Merge"
            elif new_mode == 4:
                temp_mode = 60
                n_mode_info = "Split"
            elif new_mode == 5:
                temp_mode = 36
                n_mode_info = "Lighten Only"
            elif new_mode == 6:
                temp_mode = 55
                n_mode_info = "Luma Luminance Lighten"
            elif new_mode == 7:
                temp_mode = 31
                n_mode_info = "Screen"
            elif new_mode == 8:
                temp_mode = 42
                n_mode_info = "Dodge"
            elif new_mode == 9:
                temp_mode = 33
                n_mode_info = "Addition"
            elif new_mode == 10:
                temp_mode = 35
                n_mode_info = "Darken only"
            elif new_mode == 11:
                temp_mode = 54
                n_mode_info = "Luma Luminance darken"
            elif new_mode == 12:
                temp_mode = 30
                n_mode_info = "Multiply"
            elif new_mode == 13:
                temp_mode = 43
                n_mode_info = "Burn"
            elif new_mode == 14:
                temp_mode = 53
                n_mode_info = "Linear burn"
            elif new_mode == 15:
                temp_mode = 23
                n_mode_info = "Overlay"
            elif new_mode == 16:
                temp_mode = 45
                n_mode_info = "Soft light"
            elif new_mode == 17:
                temp_mode = 44
                n_mode_info = "Hard light"
            elif new_mode == 18:
                temp_mode = 48
                n_mode_info = "Vivid light"
            elif new_mode == 19:
                temp_mode = 49
                n_mode_info = "Pin light"
            elif new_mode == 20:
                temp_mode = 50
                n_mode_info = "Linear light"
            elif new_mode == 21:
                temp_mode = 51
                n_mode_info = "Hard mix"
            elif new_mode == 22:
                temp_mode = 32
                n_mode_info = "Difference"
            elif new_mode == 23:
                temp_mode = 52
                n_mode_info = "Exclusion"
            elif new_mode == 24:
                temp_mode = 34
                n_mode_info = "Subtract"
            elif new_mode == 25:
                temp_mode = 46
                n_mode_info = "Grain extract"
            elif new_mode == 26:
                temp_mode = 47
                n_mode_info = "Grain merge"
            elif new_mode == 27:
                temp_mode = 41
                n_mode_info = "Divide"
            elif new_mode == 28:
                temp_mode = 37
                n_mode_info = "Hue HSV"
            elif new_mode == 29:
                temp_mode = 38
                n_mode_info = "Saturation HSV"
            elif new_mode == 30:
                temp_mode = 39
                n_mode_info = "Color HSL"
            elif new_mode == 31:
                temp_mode = 40
                n_mode_info = "Value HSL"
            elif new_mode == 32:
                temp_mode = 24
                n_mode_info = "Hue LCH"
            elif new_mode == 33:
                temp_mode = 25
                n_mode_info = "Chroma LCH"
            elif new_mode == 34:
                temp_mode = 26
                n_mode_info = "Color LCH"
            elif new_mode == 35:
                temp_mode = 27
                n_mode_info = "Lightness LCH"
            elif new_mode == 36:
                temp_mode = 56
                n_mode_info = "Luminance"
            mode = temp_mode
        else:
            if old_mode == 1:
                temp_mode = 1
                l_mode_info = "Dissolve"
            elif old_mode == 2:
                temp_mode = 10
                l_mode_info = "Lighten"
            elif old_mode == 3:
                temp_mode = 4
                l_mode_info = "Screen"
            elif old_mode == 4:
                temp_mode = 16
                l_mode_info = "Dodge"
            elif old_mode == 5:
                temp_mode = 7
                l_mode_info = "Addition"
            elif old_mode == 6:
                temp_mode = 9
                l_mode_info = "Darken"
            elif old_mode == 7:
                temp_mode = 3
                l_mode_info = "Multiply"
            elif old_mode == 8:
                temp_mode = 17
                l_mode_info = "Burn"
            elif old_mode == 9:
                temp_mode = 5
                l_mode_info = "Overlay"
            elif old_mode == 10:
                temp_mode = 19
                l_mode_info = "Soft light"
            elif old_mode == 11:
                temp_mode = 18
                l_mode_info = "Hard light"
            elif old_mode == 12:
                temp_mode = 6
                l_mode_info = "Difference"
            elif old_mode == 13:
                temp_mode = 8
                l_mode_info = "Subtract"
            elif old_mode == 14:
                temp_mode = 20
                l_mode_info = "Grain extract"
            elif old_mode == 15:
                temp_mode = 21
                l_mode_info = "Grain merge"
            elif old_mode == 16:
                temp_mode = 22
                l_mode_info = "Divide"
            elif old_mode == 17:
                temp_mode = 11
                l_mode_info = "Hue HSV"
            elif old_mode == 18:
                temp_mode = 12
                l_mode_info = "Saturation HSV"
            elif old_mode == 19:
                temp_mode = 13
                l_mode_info = "Color HSL"
            elif old_mode == 20:
                temp_mode = 14
                l_mode_info = "Value HSV"
            else:
                temp_mode = 0
                l_mode_info = "Normal"
            mode = temp_mode
			

        return mode, n_mode_info, l_mode_info
	
######################################################################################################################################


	
def proc_trim_away(img, layer, amount):    
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	

######################################################################################################################################

register(
    "python_fu_Custom_Font_Lighting_Effects_GT_Style_v0-05",
    "Adds lighting plug-in effects to custom fonts",
    "Provides options for adding batch lighting plug-in effects to custom fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2019",
    "<Image>/Filters/Custom Font Tools/Custom Font Lighting Effects GT Style v0.05...",
    "*",
    [
    (PF_TOGGLE, "pf_light_col", "Light colour from foreground colour (Y)\nWhite light (N): ", 0),
    #(PF_OPTION, "pf_light_type", "Type of light source: ", 1,("Point","Directional")),
    (PF_OPTION, "pf_compass", "Point light position or\nlight direction as compass point: ", 7,("N","NE","E","SE","S","SW","W","NW")),
    (PF_SPINNER, "pf_light_amount", "Distance of light source\nfrom layer centre: ", 1.50,(0.00,1000.00,0.10)),
    (PF_SPINNER, "pf_light_z", "Elevation of light source: ", 25.00,(0.10,10000.00,0.10)),
    (PF_SPINNER, "pf_intensity", "Intensity of light outside the area \nlit by the light source: ", 0.10,(0.00,100.00,0.1)),
    (PF_SPINNER, "pf_Glow", "Glow - amount of image to show\nwhere no direct light falls: ", 0.20,(0.00,100.00,0.1)),
    #(PF_SPINNER, "pf_brightness", "Brightness - intensity of light\nin area of light source: ", 0.50,(0.00,100.00,0.1)), # currently these values have no effect in non-interactive mode
    (PF_SPINNER, "pf_shininess", "Shiny - intensity of highlight: ", 2.00,(0.00,100.00,0.1)),
    (PF_SPINNER, "pf_polish", "Polished - focus of the highlight: ", 30.00,(0.00,500.00,5.00)),
    (PF_OPTION, "pf_bump_colour", "Shade of grey to use\nwith character in bumpmap layer: ", 4,("Black","Grey 1","Grey 2", "Grey 3", "Grey 4", "Grey 5", "Grey 6", "Grey 7", "Grey 8", "Grey 9", "Grey 10", "Grey 11", "Grey 12", "Grey 13","Grey 14","White" )),
    (PF_OPTION, "pf_bump_bg_colour", "Background colour to use\nfor character in bumpmap layer: ", 0,("White","Grey 14",  "Grey 13", "Grey 12", "Grey 11", "Grey 10", "Grey 9", "Grey 8", "Grey 7", "Grey 6", "Grey 5", "Grey 4", "Grey 3","Grey 2","Grey 1", "Black")),
    (PF_TOGGLE, "pf_dble_bump", "Apply environment layer as bumpmap for character: ", 1),
    (PF_TOGGLE, "pf_fixed_position", "Create random positions for the\ncharacters within the environment layer: ", 0),
    (PF_OPTION, "pf_bumpbeforeblur", "Applying blur to environment layer: ", 1,("Blur - bumpmap - blur","Bumpmap - blur","Blur - bumpmap")),
    (PF_SPINNER, "pf_bump_depth", "Bumpmap depth: ", 5.00,(0.01,65.00,0.01)),
    (PF_SPINNER, "pf_bump_blur", "Amount of blur to use with character bumpmap: ", 10.00,(0.50,100.00,1.00)),
    (PF_OPTION, "pf_bump_fill_type", "Fill method for character: ", 0, ("Fill", "Linear Gradient","Radial Gradient","Square Gradient","Conical Symmetric Gradient","Conical Asymmetric Gradient"  )),
    (PF_TOGGLE, "pf_toggle_lens", "Apply a lens to environment layer: ", 1),
    (PF_SPINNER, "pf_refraction", "Lens refraction index: ", 1.50,(1.00,100.00,0.5)),
    (PF_TOGGLE, "pf_toggle_polar", "Apply polar coordinates to environment layer: ", 0),
    (PF_OPTION, "mode_option", "Legacy Mode to use \nwith environment layer:", 0, ("Normal", "Dissolve", "Lighten", "Screen", "Dodge", "Addition", "Darken", "Multiply", "Burn", "Overlay", "Softlight", "Hardlight", "Difference", "Subtract", "Grain Extract", "Grain Merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSV)")),
    (PF_OPTION, "mode_option2", "New Mode to use \nwith environment layer:", 26, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance")),
    (PF_SPINNER, "pf_blur_radius", "Artifact blur radius: ", 10,(0.0,100.0,0.1)),
    (PF_SPINNER, "pf_blur_delta", "Artifact blur delta amount: ", 150,(0,255,1)),
    (PF_TOGGLE, "pf_replace", "Overlay existing characters with effects = Yes: \nCreate new characters with effects = No: ", 0),
    (PF_SPINNER, "pf_opacity", "Opacity of overlay: ", 50,(5,100,1)),
    (PF_TOGGLE, "pf_original", "Force original outlines: ", 0),
    ],
    [],
    do_stuff)

main()

	

