#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2019 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: Filters/Custom Font Tools/Custom Font Paths to Layers vX...
# It uses named character shaped paths to stroke and create Gimp custom fonts
# The name format consists of the required character plus a space plus the four letters char eg in quotes 'A char', 'B char' etc
# If a character path is followed by another path named 'A part', 'B part' etc that path will be stroked with a differently sized line or brush
# Character layers are generated with the usual name convention using the '.png' suffix
# This can be useful when creating certain punctuation characters such as a comma
# The plug-in will take the layer height from the parent image and attempt to size the character layer width according to the current brush or line size in the filter dialog
# Padding space can be added to the left and right margins of the characters
# The brush option provides the opportunity to create a wide range of effects - although the dialog has an option to alter brush spacing, more variation to stroking can be added 
# by the use of dynamics and animated brushes
# The line option provides a means of estimating the most effective size of brush to work with 
# Existing character layers can be re-stroked with different brushes to create a layered effect
# A space character will be created automatically with the width chosen in the dialog as 'space.png'

# Current version number: 0.03

# Version 0.02 Improved handling of characters created from two paths; catches mis-placed part paths in path stack
# Version 0.03 Further improved handling of characters created from two paths; Added options to remove paths from the file ready for saving as a custom font file

from gimpfu import *
#from array import array
import math
import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)



def do_stuff(image, layer, path_style, brush_width_major, brush_width_minor, brush_spacing, line_width_major, line_width_minor, cap_type, join_type, left_margin, right_margin, space_width, repeat_process, remove_paths, remove_all_paths):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	

    # set stroke method to paint line as a line or with the currently selected brush
    if path_style:
        stroke_method = 0
    else:
        stroke_method = 1
		
    pdb.gimp_context_set_stroke_method(stroke_method) # stroke with brush or line
	
    # Set various line stroke options
    pdb.gimp_context_set_line_width(line_width_major)
    pdb.gimp_context_set_line_join_style(join_type)
    pdb.gimp_context_set_line_cap_style(cap_type)
	
	# set colour for stroking 
    stroke_colour = pdb.gimp_context_get_foreground()
    pdb.gimp_context_set_foreground(stroke_colour)
	
    # set brush options
    pdb.gimp_context_set_sample_transparent(True)
    current_brush = pdb.gimp_context_get_brush()
    pdb.gimp_context_set_brush_size(brush_width_major)
    if brush_spacing>0: pdb.gimp_context_set_brush_spacing(float(brush_spacing/100))
    # get lists of vectors and layers ids in this image
    num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
    layer_list = image.layers
    
    # create a list of the actual layer names
    loop = 0
    while loop<len(layer_list):
        pdb.gimp_item_set_visible(layer_list[loop], False)
        layer_name = pdb.gimp_item_get_name(layer_list[loop])
        layer_name = layer_name.decode('utf-8')
        if loop == 0:
            layer_names_list = [layer_name]
        else:
            layer_names_list.append(layer_name)
        loop += 1
    found_valid_vector = False
    found_vector = False
    if num_vectors>0: # if there are any vectors in this image do something
        found_vector = True
        loop = 0
        update_canvas_size = False
        previous_name = "undefined"
        while loop<num_vectors:
            vector_id = vector_list[loop]
            active_vector = pdb.gimp_image_set_active_vectors(image,vector_id)
            active_vector_name = pdb.gimp_item_get_name(vector_id)
            active_vector_name = active_vector_name.decode('utf-8')
            if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name) and previous_name == active_vector_name[0]): # check for compatible paths
                pdb.gimp_context_set_line_width(line_width_major)
                pdb.gimp_context_set_brush_size(brush_width_major)
                found_valid_vector = True
				
                if loop<(num_vectors-1):
                    next_vector_id = vector_list[loop + 1]
                    next_vector_name = pdb.gimp_item_get_name(next_vector_id)
                    next_vector_name = next_vector_name.decode('utf-8')
                else:
                    next_vector_id = -1
                    next_vector_name = "undefined"
                #pdb.gimp_message("next_vector_name = " + next_vector_name)
				
                if " char" in active_vector_name:
                    new_name = active_vector_name[0]
                    previous_name = new_name
                    new_name = new_name.decode('utf-8')
                    new_name += ".png"
					
                    
                    layer_exists = False
                    if new_name in layer_names_list: #check to see if there is already a layer of this name
                        # and get its layer id
                        existing_layer = pdb.gimp_image_get_layer_by_name(image, new_name)
                        layer_exists = True
                    
                    # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                    maxX,minX,stroke_loop = -3000000,3000000,0
                    while stroke_loop<NumberOfStrokes_in_Path:
                        Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                        TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                        ploop,cloop = 0,0
                        #make a copy of the returned list of control points
                        Temp_ListOfControlPoints = list(ListOfControlPoints)
                        # remove unnecessary control handle entries retaining x coordinate positions for points
                        while ploop<len(ListOfControlPoints):
                            if ploop == 2 or ((ploop-2)%6 == 0):
                                cloop -= 0
                            else:
                                list_pos = ploop - cloop
                                del Temp_ListOfControlPoints[list_pos]
                                cloop += 1
                            ploop+=1
							
                        # get maximum x position from this list to find layer width
                        if max(Temp_ListOfControlPoints) > maxX: maxX = max(Temp_ListOfControlPoints)
                        # get minimum x position to use to help set layer x offset
                        if min(Temp_ListOfControlPoints) < minX: minX = min(Temp_ListOfControlPoints)
                        stroke_loop += 1
						
                    # use line width or brush width to get a path width according to type of current stroke method
                    if stroke_method == 1:
                        stroke_width = int(brush_width_major/2)
                    else:
                        stroke_width = int(line_width_major/2)
							
                    # set the width of the layer                        
                    layer_left_margin = minX - ((stroke_width) + left_margin) # define x pos of left edge of layer
                    layer_right_margin = maxX + ((stroke_width) + right_margin) # define x pos of right edge of layer
                    y_offset, char_x_offset = 0, 0

                    if layer_left_margin<0:
                        layer_width = abs(layer_left_margin - layer_right_margin)
                        char_x_offset = (layer_left_margin)
                    else:
                        layer_width = layer_right_margin - layer_left_margin
                        char_x_offset = (layer_left_margin)
							
							
                    new_layer = pdb.gimp_layer_new(image, layer_width, image.height, 1, new_name,100,0)
                    pdb.gimp_image_insert_layer(image, new_layer, None, 0)
                    if layer_width>new_layer.width:
                        pdb.gimp_layer_resize(new_layer, layer_width, image.height, 0, 0)
                        if layer_width>image.width: update_canvas_size = True
							
						
                    # move layer to stroke position
                    pdb.gimp_item_transform_translate(new_layer, char_x_offset, y_offset)


                    #pdb.gimp_message("char " + active_vector_name + ": minX = " + str(minX) + " maxX = " + str(maxX) + " layer_left_margin = " + str(layer_left_margin) + " layer_right_margin = " + str(layer_right_margin) + " layer_width = " + str(layer_width) + " char layer x offset = " + str(char_x_offset) )
                elif " part" in active_vector_name:
                    # check to see if this is a part layer to stroke in the secondary size
                    new_name = active_vector_name[0] + ".png"
                    part_name = active_vector_name[0]
                    new_name = new_name.decode('utf-8')
                    if not layer_exists: new_layer = pdb.gimp_image_get_layer_by_name(image, new_name)
                    if " part" in active_vector_name:
                        # if the path is one of ours set brush/line size
                        pdb.gimp_context_set_line_width(line_width_minor)
                        pdb.gimp_context_set_brush_size(brush_width_minor)
						
                        # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                        part_maxX,part_minX,stroke_loop = -3000000,3000000,0
                        NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                        while stroke_loop<NumberOfStrokes_in_Path:
                            Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                            TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                            ploop,cloop = 0,0
                            #make a copy of the returned list of control points
                            Temp_ListOfControlPoints = list(ListOfControlPoints)
                            # remove unnecessary control handle entries and retain x coordinate positions for points
                            while ploop<len(ListOfControlPoints):
                                if ploop == 2 or ((ploop-2)%6 == 0):
                                    cloop -= 0
                                else:
                                    list_pos = ploop - cloop
                                    del Temp_ListOfControlPoints[list_pos]
                                    cloop += 1
                                ploop+=1
                            # get maximum x position to use to help set layer width
                            if max(Temp_ListOfControlPoints) > part_maxX: part_maxX = max(Temp_ListOfControlPoints)
                            # get minimum x position to use to help set layer x offset
                            if min(Temp_ListOfControlPoints) < part_minX: part_minX = min(Temp_ListOfControlPoints)
                            stroke_loop += 1
							
                        # use line width or brush width to stroke a path
                        if stroke_method == 1:
                            stroke_width = int(brush_width_minor/2)
                        else:
                            stroke_width = int(line_width_minor/2)

                        # check if the part layer will need a wider layer than the parent character layer
                        # set the margin size for the part layers
                        part_layer_right_margin = part_maxX + ((stroke_width) + right_margin) # define x pos of right edge of part layer
                        part_layer_left_margin = part_minX - ((stroke_width) + left_margin) # define x pos of left edge of part layer
                        part_width = layer_width # provisional width
                        
                        if min(layer_left_margin, part_layer_left_margin)<0:
                            part_width = abs(min(layer_left_margin, part_layer_left_margin) - max(layer_right_margin,part_layer_right_margin) )
                        else:
                            part_width = max(layer_right_margin,part_layer_right_margin) - min(layer_left_margin, part_layer_left_margin)
                        # set value for this part layer offset
                        part_x_offset = 0
                        if part_layer_left_margin<layer_left_margin:
                            part_x_offset = abs(part_layer_left_margin - layer_left_margin)
                        else:
                            part_x_offset = 0
							
                        if part_width>layer_width:
                            # resize the character layer if extra width is required
                            pdb.gimp_layer_resize(new_layer, part_width, image.height,part_x_offset, y_offset)
                        pdb.gimp_item_transform_translate(new_layer, char_x_offset , y_offset) # move the layer into position for stroking
						
                        if part_width>image.width: update_canvas_size = True # set flag for updating canvas size
													
                # make layer visible
                pdb.gimp_item_set_visible(new_layer, True)
					
                # stroke the path (using current width/colour settings) the required number of times
                repeat_loop = 0
                while repeat_loop<repeat_process:
                    pdb.gimp_drawable_edit_stroke_item(new_layer, vector_id)
                    pdb.gimp_displays_flush()
                    repeat_loop += 1
					
					
                # reset the layer position
                new_layer.set_offsets(0, 0)
				
				
                if "part" not in next_vector_name and layer_exists:
                    pdb.gimp_item_set_visible(new_layer, True)
                    pdb.gimp_item_set_visible(existing_layer, True)
                    pdb.gimp_image_merge_down(image, new_layer, 0)
                    new_layer = image.active_layer 
						
				
                # turn off layer visibility
                if loop <> (num_vectors-1): pdb.gimp_item_set_visible(new_layer, False) # turn off layer visibility
                pdb.gimp_displays_flush()
            loop +=1
		
        if update_canvas_size: pdb.gimp_image_resize_to_layers(image)
		
        # check to see if there is an existing space layer and create one if necessary
        if not "space.png"  in layer_names_list and found_valid_vector == True:
            space_width = (float(image.width) / 100) * space_width
            space_layer = pdb.gimp_layer_new(image,(space_width), image.height, 1, "space.png",100,0)
            pdb.gimp_image_insert_layer(image, space_layer, None, 0)
            # make space layer not visible
            pdb.gimp_item_set_visible(space_layer, False) # turn off layer visibility
        if found_valid_vector: pdb.gimp_image_set_active_layer(image, new_layer)
        if found_valid_vector: pdb.gimp_item_set_visible(new_layer, True) # turn off layer visibility				
			
        if found_valid_vector == False: # Report no suitable paths
            old_handler = pdb.gimp_message_get_handler()		
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("This image has at least one path but there are none suitable for the plug-in to work with.")	
            pdb.gimp_message_set_handler(old_handler)
        else:
            if remove_paths or remove_all_paths:
                # get lists of vector ids in this image
                num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
                vector_list = image.vectors
                loop = 0
                while loop<num_vectors:
                    vector_id = vector_list[loop]
                    active_vector_name = pdb.gimp_item_get_name(vector_id)
                    active_vector_name = active_vector_name.decode('utf-8')
                    if not remove_all_paths:
                        if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name)): 
                            pdb.gimp_image_remove_vectors(image, vector_list[loop])
                    else:
                        pdb.gimp_image_remove_vectors(image, vector_list[loop])
                    loop += 1
                
	
    if found_vector == False: # Report no paths
            old_handler = pdb.gimp_message_get_handler()		
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("There are no paths, associated with this image.")	
            pdb.gimp_message_set_handler(old_handler)	
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

   

######################################################################################################################################
	
register(
    "python_fu_Paths_To_Layers_v0-03",
    "Provides options for turning paths into character layers ",
    "Provides options for turning paths into character layers ",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2019",
    "<Image>/Filters/Custom Font Tools/Custom Font Paths to Layers v.0.03...",
    "RGBA",
    [
    (PF_OPTION, "option", "Stroke path with: ", 0, ("Brush", "Line")),
    (PF_SPINNER, "brush_primary", "Brush size (primary strokes):", 90, (10, 2000, 1)),
    (PF_SPINNER, "brush_secondary", "Brush size (secondary strokes):", 45, (10, 2000, 1)),
    (PF_SPINNER, "brush_spacing", "Brush spacing:", 50, (0, 5000, 1)),
    (PF_SPINNER, "line_width_primary", "Path line width (primary strokes):", 90, (1, 2000, 1)),
    (PF_SPINNER, "line_width_secondary", "Path line width (secondary strokes):", 45, (1, 2000, 1)),
    (PF_OPTION, "cap_option", "Path end cap style: ", 1, ("Butt", "Round", "Square")),
    (PF_OPTION, "join_option", "Path line join style: ", 1, ("Mitre", "Round", "Bevel")),
    (PF_SPINNER, "left_margin", "Left hand character margin adjustment:", 10, (-500, 500, 1)),
    (PF_SPINNER, "right_margin", "Right hand character margin adjustment:", 10, (-500, 500, 1)),
    (PF_SPINNER, "space_size", "Width of space character (as % of image width):", 50, (10, 100, 1)),
    (PF_SPINNER, "repeats", "Repeat strokes:", 3, (1, 50, 1)),
    (PF_TOGGLE, "path_option", "Remove character paths only \nfrom custom font file:", 0),
    (PF_TOGGLE, "path_all_option", "Remove all paths \nfrom custom font file:", 0),
    ],
    [],
    do_stuff)

main()
