#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2020 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Single Path Stroker
#
# Version 001 Provides options for stroking paths with lines (or the currently selected brush)
#             Path width option (also controls number of colours sampled)
#             Style options for joins and end caps
#             Brush options to control number of times paths are stroked
#             Glow options - colour/amount/constrain
#             Left/Right margin adjustments
#             Space width option
#             Remove path option

# Version 002/002b Added options for varying the width of the stroke of nominated characters. Currently two extra sizes can be used.
#             The groups are created by adding text layers containing the nominated characters and renaming them as thinner and thinnest.
#             Obviously you can use both or just one or no groups. The layers are deleted if the option to delete character paths is set.
#             This option is intended to help make smaller characters more legible particularly when using a brush to stroke wider paths.
#             This version (0.02b) allows you to group the characters to apply these configurable stroke widths to. Unfortunately some
#             characters cannot be used directly in code to show in the filter dialog - mainly the non-keyboard characters - and so
#             I have used a less transparent mechanism. This is fine if you are happy with my character groupings or if you are happy to
#             insert your own characters each time you use the plug-in. 
#             If you want to make permanent changes to the groupings use the lists below as a reference for adding non-keyboard characters. 
#             
#             Use your 'thinner' characters at line 702 and remove the hash character from the start of the line
#             Add a hash character to comment out my line (line 703) and save the file. Close Gimp, if it is running, and then start up again.
#             Edit and/or add keyboard characters between the double quotes. Add non-keyboard characters by adding the following code snippet 
#             immediately after the second double quote +str(unichr(XXX))
#             Replace XXX with the code for your chosen character code.
#             If the filter does not show up in the menu and you cannot see where you have gone wrong, put the hash character back on 
#             at line 679 and remove the hash from line 678 to repair the filter. Repeat this process for the thinnest string if necessary.
# Version 003 Only continuing with the version without added thinner/thinnest text layers (due to fewer downloads)
#             Changed unicode character value to 8364 for Euro currency symbol
#             Added error message for when paths are detected but there are no correctly named paths

#             ##### COMMON KEYCODE CHARACTER REFERENCE GUIDE ####
#             Code Snippet					Character
#             +str(unichr(34))			 	"		double quote character
#             +str(unichr(8364))			Euro currency character
#             +str(unichr(162))			 	cent character
#             +str(unichr(163))			 	pound sterling character
#             +str(unichr(169))			 	copyright symbol
#             +str(unichr(174))			 	registered symbol
#             +str(unichr(176))			 	degree symbol
#             +str(unichr(178))				superscript 2
#             +str(unichr(179))			 	superscript 3
#             +str(unichr(183))			 	middle dot/decimal point
#             +str(unichr(185))			 	superscript 1
#             +str(unichr(188))			 	fraction one quarter
#             +str(unichr(189))			 	fraction one half
#             +str(unichr(190))			 	fraction three quarters
#             +str(unichr(215))			 	multiplication sign
#             +str(unichr(247))			 	division sign
#             

from gimpfu import *
#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)



def do_stuff(image, layer, stroke_width, colour_style, cap_style, join_style, stroke_method, brush_options, glow_source, blur_amount, do_constrain, left_margin, right_margin, thinner_chars, thinner_seed, thinnest_chars, thinnest_seed, space_width, remove_paths):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
    merge_layers = True # leftover option
    original_stroke_width = stroke_width
    initial_stroke_width = stroke_width
    active_path = pdb.gimp_image_get_active_vectors(image)
	
    paths_to_use = False
	
    num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
    
    # look for at least one correctly named path
    loop = 0
    while loop<len(vector_list):
        vector_name = pdb.gimp_item_get_name(vector_list[loop])
        vector_name = vector_name.decode('utf-8') # in case of unicode characters 
        if (len(vector_name) == 6 and " char" in vector_name) or len(vector_name) == 1: 
            paths_to_use = True
            loop = len(vector_list)
        loop += 1
	
    #pdb.gimp_message("paths_to_use = " + str(paths_to_use) )
	
    if active_path and paths_to_use:
	
        pdb.gimp_image_resize(image, image.width * 1.5, image.height,0,0) # quick and dirty way to ensure that no layers (hopefully) are cropped on the right-hand side of their layers
        # get paths/vectors/layers ids
        num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        layer_list = image.layers
		
        # create a list of the actual layer names
        loop = 0
		
        thinner, thinnest, thinner_size, thinnest_size, thinner_id, thinnest_id = "", "", 0, 0, -1, -1
        thinner,thinnest = thinner_chars, thinnest_chars
        while loop<len(layer_list):
            pdb.gimp_item_set_visible(layer_list[loop], False)
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8') # in case of unicode characters 
            # commented code below used when text layers used to modify path widths for groups of characters
            #if layer_name.lower() == "thinner": # check to see if there is a text layer containing characters to be stroked at a thinner width
            #    if pdb.gimp_item_is_text_layer(layer_list[loop]):
            #        thinner = pdb.gimp_text_layer_get_text(layer_list[loop])
            #    thinner_id = layer_list[loop]
            #if layer_name.lower() == "thinnest":  # check to see if there is a text layer containing characters to be stroked at the thinnest width
            #    if pdb.gimp_item_is_text_layer(layer_list[loop]):
            #        thinnest = pdb.gimp_text_layer_get_text(layer_list[loop])
            #    thinnest_id = layer_list[loop]
            if loop == 0:
                layer_names_list = [layer_name]
            else:
                layer_names_list.append(layer_name)
            loop += 1
			
        # set the stroke width size from the user dialog values
        if thinner <> "": 
            thinner_size = int((stroke_width/100) * thinner_seed)
        if thinnest <> "":
            thinnest_size = int((stroke_width/100) * thinnest_seed)
			
			
        # create a list of the actual vector names
        loop, part_flag = 0, False
		
        while loop<len(vector_list):
            pdb.gimp_item_set_visible(vector_list[loop], False)
            vector_name = pdb.gimp_item_get_name(vector_list[loop])
            vector_name = vector_name.decode('utf-8') # in case of unicode characters 
            if "part" in vector_name: part_flag = True
            if loop == 0:
                vector_names_list = [vector_name]
            else:
                vector_names_list.append(vector_name)
				
            loop += 1
			
        if part_flag: 
            # Since there is no pdb to merge two paths together we need to 
            # deal with character path files that use two character parts i.e. one ending in char that acts as the base character 
            # and one ending in part by joining the part vector onto the char vector
            loop = 0
            while loop<len(vector_list):
                vector_name = vector_names_list[loop]
                vector_id = vector_list[loop]
                if (len(vector_name) == 6 and " char" in vector_name): #check that the vector is one we can use
                    next_vector_name = "none"
                    if loop < (len(vector_list)-1):
                        next_vector_name = vector_names_list[loop+1]
                        #pdb.gimp_message(next_vector_name)
                        if (len(next_vector_name) == 6 and " part" in next_vector_name): # look to see if the next vector needs joining onto the it
                            part_vector_id = vector_list[loop + 1]
                            # get paths etc associated with the part vector
                            num_part_strokes, part_vector_list = pdb.gimp_vectors_get_strokes(part_vector_id)
                            stroke_loop = 0
                            
                            while stroke_loop<num_part_strokes:
                                # loop through paths that make up the part vector
                                part_vector_stroke = part_vector_list[stroke_loop]
                                # get information associated with the current vector part
                                TypeOfStrokeInPart_Vector,NumberOfPointsInPart_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(part_vector_id, part_vector_stroke)
                                part_ListOfControlPoints = ListOfControlPoints
                                # add current part path details to the base character
                                vector_id = pdb.gimp_vectors_stroke_new_from_points(vector_list[loop], 0, NumberOfPointsInPart_Vector, ListOfControlPoints, IsStrokeClosed)
							
                                stroke_loop += 1
				
				
                loop += 1
			
        

        # ensure that stroke width is a multiple of 2
        if stroke_width>3:
            if stroke_width % 2 == 1: stroke_width += 1
            number_of_strokes = (stroke_width/2) - 1
        else:
            number_of_strokes = 1
			
        # detect current colours
        original_fg_colour = pdb.gimp_context_get_foreground()
        original_bg_colour = pdb.gimp_context_get_background()
        original_current_gradient = pdb.gimp_context_get_gradient()
        current_gradient = original_current_gradient
        # get colour information
        list_of_gradient_positions = fn_get_gradient_positions(number_of_strokes)
        number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
        update_canvas_size = False
		
        path_loop = 0
        while path_loop < num_paths:
            stroke_width = initial_stroke_width # begin each character by setting the width to its maximum size
            # look for recognised path
            path_name = pdb.gimp_item_get_name(vector_list[path_loop])
            path_name = path_name.decode('utf-8')
            if len(path_name) == 1 or (len(path_name) == 6 and " char" in path_name):
                layer_name = path_name[0] + ".png"
				
                # set the stroke width if the character is in the appropriate list
                if path_name[0] in thinner:
                    initial_stroke_width = thinner_size
                    stroke_width = thinner_size
                elif path_name[0] in thinnest:
                    initial_stroke_width = thinnest_size
                    stroke_width = thinnest_size
                else:
                    initial_stroke_width = original_stroke_width
                    stroke_width = original_stroke_width
	
                if stroke_width>3:
                    if stroke_width % 2 == 1: stroke_width += 1
                    number_of_strokes = (stroke_width/2) - 1
                else:
                    number_of_strokes = 1
					
                # reset colours for first stroke along path
                fg_colour = original_fg_colour
                bg_colour = original_bg_colour
                pdb.gimp_context_set_foreground(fg_colour)
                pdb.gimp_context_set_background(bg_colour)
               
                current_gradient = original_current_gradient
                # get colour information
                list_of_gradient_positions = fn_get_gradient_positions(number_of_strokes)
                number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
			
                pdb.gimp_context_set_stroke_method(stroke_method)
                if stroke_method == 1:
                    current_brush = pdb.gimp_context_get_brush() # get current brush
                    if current_brush == "Clipboard Image":
                        if not use_clipboard_as_brush: # we do not want to use the clipboard image as a brush
                            number_of_brushes, brushes_list = pdb.gimp_brushes_get_list("")
                            if number_of_brushes > 2:
                                pdb.gimp_context_set_brush(brushes_list[2])
						
                if stroke_method == 1: # stroke path with brush
                    pdb.gimp_context_set_brush_size(stroke_width*1.75) # set brush size
                else: # set line options
                    pdb.gimp_context_set_line_width(float(stroke_width))
                    pdb.gimp_context_set_line_cap_style(cap_style)
                    pdb.gimp_context_set_line_join_style(join_style)
			
                pdb.gimp_context_set_stroke_method(stroke_method)
                
                # calculate width of layer 
                width, char_x_offset, y_offset, update_canvas_size = fn_calculate_layer_width(vector_list[path_loop], stroke_width, left_margin, right_margin, blur_amount, image, update_canvas_size)
                height = image.height
				
                loop, fib1, fib2 = 0, 0, 1
                while loop < number_of_strokes:
            	
				
                    new_layer = gimp.Layer(image, "Path Layer", width, height, RGBA_IMAGE, 100, 0) # define new layer id
                    image.add_layer(new_layer, -1) # and add new layer to image
					
					
                    if width>new_layer.width: # resize layer if necessary
                        pdb.gimp_layer_resize(new_layer, layer_width, image.height, 0, 0)
                        if width>image.width: update_canvas_size = True
						
						
                    # move layer to stroke position
                    pdb.gimp_item_transform_translate(new_layer, char_x_offset, y_offset)
					
                    if colour_style == 2: # set colour options
                        path_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                    elif colour_style == 0:
                        path_colour = fn_shade_text_colour(fg_colour, bg_colour, number_of_strokes, loop)
                    else:
                        path_colour = fn_shade_text_colour(bg_colour, fg_colour, number_of_strokes, loop)
			
                    pdb.gimp_context_set_foreground(path_colour)
					
					
                    if stroke_method == 1: 
                        pdb.gimp_context_set_brush_size(stroke_width*1.75)
                    else:
                        
                        if stroke_width<1: stroke_width = 1
                        pdb.gimp_context_set_line_width(float(stroke_width))
						
                    pdb.gimp_drawable_edit_stroke_item(new_layer, vector_list[path_loop]) # stroke the path
					
                    if loop == 0: pdb.plug_in_gauss(image, new_layer,float(1),float(1),0) # blur first path
					
                    if (loop>0 and merge_layers == True): new_layer = pdb.gimp_image_merge_down(image, new_layer, 0) # merge subsequent layers
					
                    stroke_width -= 2 # reduce width of next stroke
                    success = pdb.gimp_displays_flush()
					
							
							
                    if brush_options <2:
                        loop += 1
                    elif brush_options == 2:
                        loop = number_of_strokes
                    elif brush_options == 3:
                        loop = fib1 + fib2
                        fib1 = fib2
                        fib2 = loop
										
							
                num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
                vector_list = image.vectors
			
                if glow_source>0: # apply a glow effect if required
                    blur_loop = 0
                    blur = blur_amount	
			
                    if glow_source == 1: # choose colour type
                        path_colour = original_bg_colour
                    else:
                        path_colour = original_fg_colour
				
			
                    # reset the original character layer position to image origin
                    new_layer.set_offsets(0, 0)
					
                    active_path = vector_list[path_loop] # set active path
                    pdb.gimp_image_set_active_vectors(image, active_path)
            	
                    blur_layer = gimp.Layer(image, "Blur Layer", width, height, RGBA_IMAGE, 100, 0) # create and add a layer for blurring to use as a glow
                    image.add_layer(blur_layer, -1)
                    pdb.gimp_context_set_foreground(path_colour)
                    pdb.gimp_context_set_line_width(float(initial_stroke_width-1)) # reduce width of line by one pixel
                    pdb.gimp_item_transform_translate(active_path, abs(char_x_offset), y_offset) # move path to correct position
                    pdb.gimp_drawable_edit_stroke_item(blur_layer, active_path) # and stroke it
					
                    if do_constrain: 
                        pdb.gimp_image_select_item(image, 2, new_layer) # turn on constraining selection for glow
                        pdb.gimp_selection_grow(image, initial_stroke_width/2)
                        pdb.gimp_selection_feather(image, 2)
						
                    # do some blurring
                    blur = blur // 1.25
                    pdb.plug_in_gauss(image, blur_layer,float(blur),float(blur),0)
					
					
                    pdb.gimp_drawable_edit_stroke_item(blur_layer, active_path)
                    blur = blur // 1.25
                    pdb.plug_in_gauss(image, blur_layer,float(blur),float(blur),0)
					
                    if (blur_loop>0): pdb.gimp_image_merge_down(image, blur_layer, 0) # merge down subsequent blurring layers				
					
                    pdb.gimp_drawable_edit_stroke_item(blur_layer, active_path) # stroke the path
                    pdb.gimp_image_lower_item(image, blur_layer) # lower the blur so that it is below the character layer
                    new_layer = pdb.gimp_image_merge_down(image, new_layer, 0) # and merge the two layers
					
                    if do_constrain: pdb.gimp_selection_none(image) # turn off constraining selection for glow
					
                    char_width, char_height = new_layer.width, new_layer.height # get width/height of character layer
                    crop_layer = pdb.gimp_layer_copy(new_layer, True) # copy the character layer and add it as a new layer
                    image.add_layer(crop_layer, -1)
                    pdb.plug_in_autocrop_layer(image, crop_layer) # autocrop the layer to fit around the glow
                    crop_width, crop_height = crop_layer.width, crop_layer.height # get the width/height of this cropped layer
					
                    crop_x_offset, crop_y_offset = pdb.gimp_drawable_offsets(crop_layer) # get the offsets of this cropped layer
					
                    if do_constrain: # calculate new width
                        blurred_width = crop_width + left_margin + right_margin
                        new_x_offset = (0 - (crop_x_offset)) + left_margin
                    else:
                        blurred_width = crop_width
                        new_x_offset =  0 - crop_x_offset
						
                    pdb.gimp_layer_resize(crop_layer, image.width, image.height, crop_x_offset, crop_y_offset) # resize the cropped layer to the full size of the image
                    pdb.gimp_image_remove_layer(image, new_layer) # remove the old character layer
                    pdb.gimp_layer_resize(crop_layer, blurred_width, char_height, new_x_offset, 0) # resize the cropped layer to the correct size
					
					
                    active_layer = image.active_layer 
                    new_layer = active_layer # reset the layer id
				
                # reset the original character layer position to image origin
                new_layer.set_offsets(0, 0)
				
                
                pdb.gimp_item_set_name(new_layer,layer_name)
                if path_loop<(num_paths-1): pdb.gimp_item_set_visible(new_layer, False) # turn off visibility
			
            path_loop += 1
			
        pdb.gimp_image_resize_to_layers(image)
			
        # check to see if there is an existing space layer and create one if necessary
        if not "space.png"  in layer_names_list:
            space_width = (float(image.width) / 100) * space_width
            space_layer = pdb.gimp_layer_new(image,(space_width), image.height, 1, "space.png",100,0)
            pdb.gimp_image_insert_layer(image, space_layer, None, 0)
            # make space layer not visible
            pdb.gimp_item_set_visible(space_layer, False) # turn off layer visibility
            pdb.gimp_image_set_active_layer(image, new_layer)
            pdb.gimp_item_set_visible(new_layer, True) # turn off layer visibility
			

			
        if remove_paths:
            # get lists of vector ids in this image
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            loop = 0
            while loop<num_vectors:
                vector_id = vector_list[loop]
                active_vector_name = pdb.gimp_item_get_name(vector_id)
                active_vector_name = active_vector_name.decode('utf-8')
                if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name)) or (len(active_vector_name) == 1): 
                    pdb.gimp_image_remove_vectors(image, vector_list[loop])
                loop += 1
				
            if thinner_id>-1: pdb.gimp_image_remove_layer(image, thinner_id)
            if thinnest_id>-1: pdb.gimp_image_remove_layer(image, thinnest_id)
			
    else:
        if not paths_to_use:
            message_text = "Single Path Stroker cannot find a path with a suitable name.\nSingle Path Stroker uses paths named with a single character or a single character followed by a space character and the four letter suffix char."
        else:
            message_text = "There appears to be no active paths to use."
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)  
	
	
	
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################
	
def fn_calculate_layer_width(vector_id, stroke_width, left_margin, right_margin, blur_amount, image, update_canvas_size):
    # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
    maxX,minX,stroke_loop = -3000000,3000000,0
    while stroke_loop<NumberOfStrokes_in_Path:
        Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
        TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
        ploop,cloop = 0,0
        #make a copy of the returned list of control points
        Temp_ListOfControlPoints = list(ListOfControlPoints)
        # remove unnecessary control handle entries retaining x coordinate positions for points
        while ploop<len(ListOfControlPoints):
            if ploop == 2 or ((ploop-2)%6 == 0):
                cloop -= 0
            else:
                list_pos = ploop - cloop
                del Temp_ListOfControlPoints[list_pos]
                cloop += 1
            ploop+=1
							
        # get maximum x position from this list to find layer width
        if max(Temp_ListOfControlPoints) > maxX: maxX = max(Temp_ListOfControlPoints)
        # get minimum x position to use to help set layer x offset
        if min(Temp_ListOfControlPoints) < minX: minX = min(Temp_ListOfControlPoints)
        stroke_loop += 1
						
    # use line width or brush width to get a path width according to type of current stroke method
    #if stroke_method == 1:
    #    stroke_width = int(brush_width_major/2)
    #else:
    #    stroke_width = int(line_width_major/2)
							
    # set the width of the layer                        
    layer_left_margin = minX - ((stroke_width) + max(left_margin, blur_amount)) # define x pos of left edge of layer
    layer_right_margin = maxX + ((stroke_width) + max(right_margin, blur_amount)) # define x pos of right edge of layer
    y_offset, char_x_offset = 0, 0

    if layer_left_margin<0:
        layer_width = abs(layer_left_margin - layer_right_margin)
        char_x_offset = (layer_left_margin)
    else:
        layer_width = layer_right_margin - layer_left_margin
        char_x_offset = (layer_left_margin)
		
    layer_width = int(layer_width)
		
    if layer_width>image.width: update_canvas_size = True
		
    return layer_width, char_x_offset, y_offset, update_canvas_size
						
######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
#######################################################################################################

def fn_shade_gradient_colour(colours_of_samples, instances):
    red_amount = colours_of_samples[(instances*4) + 0] * 255
    green_amount = colours_of_samples[(instances*4) + 1] * 255
    blue_amount = colours_of_samples[(instances*4) + 2] * 255
    alpha_amount = colours_of_samples[(instances*4) + 3] * 255

    if red_amount < 0: red_amount = 0
    if red_amount > 255: red_amount = 255
    if green_amount < 0: green_amount = 0
    if green_amount > 255: green_amount = 255
    if blue_amount < 0: blue_amount = 0
    if blue_amount > 255: blue_amount = 255
    if alpha_amount < 0: alpha_amount = 0
    if alpha_amount > 255: alpha_amount = 255
	
    new_f_colour = (int(red_amount), int(green_amount), int(blue_amount), int(alpha_amount))
    return new_f_colour
	
#######################################################################################################

def fn_get_gradient_positions(instances):
    # get positions along gradients for text colours
    grad_position = 0.0
    if instances<1: instances = 1
    grad_amount = float(1.00/instances)
    loop = 0
    while loop<=instances:
        if loop == 0:
            list_of_pos = [grad_position]
        else:
            list_of_pos.append(grad_position)
        grad_position += grad_amount
        loop += 1
    return list_of_pos

#######################################################################################################
	
def fn_shade_text_colour(f_colour, bl_colour, instances, amount):
    red_amount_old = f_colour[0]
    green_amount_old = f_colour[1]
    blue_amount_old = f_colour[2]
    red_amount_new = bl_colour[0]
    green_amount_new = bl_colour[1]
    blue_amount_new = bl_colour[2]
    zero = 0.0
	
    if red_amount_old<red_amount_new:
        red_spread = float(red_amount_new - red_amount_old)
        red_step = float(red_spread/instances)
    else:
        red_spread = float(red_amount_old - red_amount_new)
        red_step = float(zero - (red_spread/instances))
		
    if green_amount_old<green_amount_new:
        green_spread = float(green_amount_new - green_amount_old)
        green_step = float(green_spread/instances)
    else:
        green_spread = float(green_amount_old - green_amount_new)
        green_step = float(zero - float((green_spread/instances)))

    if blue_amount_old<blue_amount_new:
        blue_spread = float(blue_amount_new - blue_amount_old)
        blue_step = float(blue_spread/instances)
    else:
        blue_spread = float(blue_amount_old - blue_amount_new)
        blue_step = float(zero - float((blue_spread/instances)))

    new_red_amount = red_amount_old + (red_step * amount)
    if new_red_amount < 0: new_red_amount = 0
    if new_red_amount > 255: new_red_amount = 255
	
    new_green_amount = green_amount_old + (green_step * amount)
    if new_green_amount < 0: new_green_amount = 0
    if new_green_amount > 255: new_green_amount = 255
	
    new_blue_amount = blue_amount_old + (blue_step * amount)
    if new_blue_amount < 0: new_blue_amount = 0
    if new_blue_amount > 255: new_blue_amount = 255
	
    new_f_colour = (int(new_red_amount), int(new_green_amount), int(new_blue_amount))
	
    return new_f_colour
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value


######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Single_Path_Stroker_v_003",
    "Provides options for stroking character paths",
    "Provides options for stroking character paths",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2020",
    "<Image>/Filters/Custom Font Tools/Custom Font Single Path Stroker v 0.03",
    "RGB*",
    [
    (PF_SPINNER, "size", "Maximum path width", 20, (12, 500, 2)),
    (PF_OPTION, "colour_option", "Take colours from: ", 2, ("Foreground - Background","Background - Foreground", "Current gradient")),
    (PF_OPTION, "pf_cap_style", "Line cap style", 1, ("Butt", "Round", "Square")),
    (PF_OPTION, "pf_join_style", "Line join style", 1, ("Mitre", "Round", "Bevel")),
    (PF_OPTION, "pf_stroke_method", "Stroke path method", 0, ("Stroke with line", "Stroke with current brush")),
    (PF_OPTION, "brush_options", "Stroke options: ", 0, ("Stroke every path", "Stroke every other path","Stroke first path only", "Stroke in Fibonacci sequence")),
    (PF_OPTION, "pf_glow_method", "Glow colour: ", 0, ("No glow", "Use background colour", "Use foreground colour")),
    (PF_SPINNER, "blur_size", "Glow dispersal:", 0, (0, 500, 10)),
    (PF_TOGGLE, "constrain_option", "Constrain glow within character:", 1),
    (PF_SPINNER, "left_margin", "Left hand character margin adjustment:", 10, (-500, 500, 1)),
    (PF_SPINNER, "right_margin", "Right hand character margin adjustment:", 10, (-500, 500, 1)),
    #(PF_STRING, "thinner_string", "Characters to stroke\nin the thinner width:", "><|}{+_)(*&^%$#@!~/\][=-"+str( unichr(8364))),
    (PF_STRING, "thinner_string", "Characters to stroke\nin the thinner width:", "><|}{+_)(*&^%$#@!~/\][=-"+str( unichr(8364))+str( unichr(162))+str( unichr(163))+str( unichr(169))+str( unichr(174))+str( unichr(188))+str( unichr(189))+str( unichr(190))+str( unichr(215))+str( unichr(247))),
    (PF_SPINNER, "thinner_size", "Size of any characters\nnominated in thinner group:", 100, (50, 100, 1)),
    #(PF_STRING, "thinnest_string", "Characters to stroke\nin the thinnest width:", ":.,';`"+str( unichr(34)))
    (PF_STRING, "thinnest_string", "Characters to stroke\nin the thinnest width:", ":.,';`"+str( unichr(34))+str( unichr(176))+str( unichr(178))+str( unichr(179))+str( unichr(183))+str( unichr(185))),
    (PF_SPINNER, "thinnest_size", "Size of any characters\nnominated in thinnest group:", 100, (25, 100, 1)),
    (PF_SPINNER, "space_size", "Width of space character (as % of image width):", 50, (10, 100, 1)),
    (PF_TOGGLE, "path_option", "Remove character paths\nfrom custom font file:", 1),
    ],
    [],
    do_stuff)

main()
